#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import os

import pytest

from molecule.config import Config
from molecule.util import write_file
from molecule.verifier import testinfra


@pytest.fixture
def _patched_testinfra_get_tests(mocker):  # type: ignore[no-untyped-def]  # noqa: ANN202
    m = mocker.patch("molecule.verifier.testinfra.Testinfra._get_tests")
    m.return_value = ["foo.py", "bar.py"]

    return m


@pytest.fixture
def _verifier_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "verifier": {
            "name": "testinfra",
            "options": {"foo": "bar", "v": True, "verbose": True},
            "additional_files_or_dirs": ["file1.py", "file2.py", "match*.py", "dir/*"],
            "env": {"FOO": "bar"},
        },
    }


# NOTE(retr0h): The use of the `patched_config_validate` fixture, disables
# config.Config._validate from executing.  Thus preventing odd side-effects
# throughout patched.assert_called unit tests.
@pytest.fixture
def _instance(patched_config_validate, config_instance: Config):  # type: ignore[no-untyped-def]  # noqa: ANN202
    return testinfra.Testinfra(config_instance)


@pytest.fixture
def inventory_file(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return _instance._config.provisioner.inventory_file


@pytest.fixture(name="inventory_directory")
def fixture_inventory_directory(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    return _instance._config.provisioner.inventory_directory


def test_testinfra_config_private_member(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert isinstance(_instance._config, Config)


def test_testinfra_default_options_property(inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {
        "connection": "ansible",
        "ansible-inventory": inventory_directory,
        "p": "no:cacheprovider",
    }

    assert x == _instance.default_options


def test_default_options_property_updates_debug(inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._config.args = {"debug": True}
    x = {
        "connection": "ansible",
        "ansible-inventory": inventory_directory,
        "debug": True,
        "vvv": True,
        "p": "no:cacheprovider",
    }

    assert x == _instance.default_options


def test_default_options_property_updates_sudo(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    inventory_directory,
    _instance,  # noqa: PT019
    _patched_testinfra_get_tests,  # noqa: PT019
):
    _instance._config.args = {"sudo": True}
    x = {
        "connection": "ansible",
        "ansible-inventory": inventory_directory,
        "sudo": True,
        "p": "no:cacheprovider",
    }

    assert x == _instance.default_options


def test_testinfra_default_env_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert "MOLECULE_FILE" in _instance.default_env
    assert "MOLECULE_INVENTORY_FILE" in _instance.default_env
    assert "MOLECULE_SCENARIO_DIRECTORY" in _instance.default_env
    assert "MOLECULE_INSTANCE_CONFIG" in _instance.default_env


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_additional_files_or_dirs_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    tests_directory = _instance._config.verifier.directory
    file1_file = os.path.join(tests_directory, "file1.py")  # noqa: PTH118
    file2_file = os.path.join(tests_directory, "file2.py")  # noqa: PTH118
    match1_file = os.path.join(tests_directory, "match1.py")  # noqa: PTH118
    match2_file = os.path.join(tests_directory, "match2.py")  # noqa: PTH118
    test_subdir = os.path.join(tests_directory, "dir")  # noqa: PTH118
    test_subdir_file = os.path.join(test_subdir, "test_subdir_file.py")  # noqa: PTH118

    os.mkdir(tests_directory)  # noqa: PTH102
    os.mkdir(test_subdir)  # noqa: PTH102
    for f in [file1_file, file2_file, match1_file, match2_file, test_subdir_file]:
        write_file(f, "")

    x = [file1_file, file2_file, match1_file, match2_file, test_subdir_file]
    assert sorted(x) == sorted(_instance.additional_files_or_dirs)


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_testinfra_env_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.env["FOO"] == "bar"
    assert "ANSIBLE_CONFIG" in _instance.env


def test_testinfra_name_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.name == "testinfra"


def test_testinfra_enabled_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.enabled


def test_testinfra_directory_property(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    parts = _instance.directory.split(os.path.sep)

    assert parts[-3:] == ["molecule", "default", "tests"]


@pytest.fixture
def _verifier_testinfra_directory_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {"verifier": {"name": "testinfra", "directory": "/tmp/foo/bar"}}  # noqa: S108


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_testinfra_directory_section_data"],  # noqa: PT007
    indirect=True,
)
def test_directory_property_overridden(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.directory == "/tmp/foo/bar"  # noqa: S108


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_testinfra_options_property(inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    x = {
        "connection": "ansible",
        "ansible-inventory": inventory_directory,
        "foo": "bar",
        "v": True,
        "verbose": True,
        "p": "no:cacheprovider",
    }

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_testinfra_options_property_handles_cli_args(inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._config.args = {"debug": True}
    x = {
        "connection": "ansible",
        "ansible-inventory": inventory_directory,
        "foo": "bar",
        "debug": True,
        "vvv": True,
        "verbose": True,
        "p": "no:cacheprovider",
    }

    assert x == _instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_testinfra_bake(_patched_testinfra_get_tests, inventory_directory, _instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    _instance._tests = ["foo.py", "bar.py"]
    _instance.bake()
    args = [
        "pytest",
        "--ansible-inventory",
        inventory_directory,
        "--connection",
        "ansible",
        "--foo",
        "bar",
        "-p",
        "no:cacheprovider",
        "foo.py",
        "bar.py",
        "-v",
    ]

    assert _instance._testinfra_command == args


def test_testinfra_execute(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    caplog,
    patched_run_command,
    _patched_testinfra_get_tests,  # noqa: PT019
    _instance,  # noqa: PT019
):
    _instance.execute()

    patched_run_command.assert_called_once()

    msg = f"Executing Testinfra tests found in {_instance.directory}/..."
    msg2 = "Verifier completed successfully."
    assert msg in caplog.text
    assert msg2 in caplog.text
    assert patched_run_command.call_args[0][0][0] == "pytest"


def test_testinfra_execute_does_not_execute(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    caplog,
    _instance,  # noqa: PT019
):
    _instance._config.config["verifier"]["enabled"] = False
    _instance.execute()

    assert not patched_run_command.called

    msg = "Skipping, verifier is disabled."
    assert msg in caplog.text


def test_does_not_execute_without_tests(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    caplog,
    _instance,  # noqa: PT019
):
    _instance.execute()

    assert not patched_run_command.called

    msg = "Skipping, no tests found."
    assert msg in caplog.text


def test_testinfra_execute_bakes(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    _patched_testinfra_get_tests,  # noqa: PT019
    _instance,  # noqa: PT019
):
    _instance.execute()

    assert _instance._testinfra_command is not None

    assert patched_run_command.call_count == 1


@pytest.mark.parametrize(
    "config_instance",
    ["_verifier_section_data"],  # noqa: PT007
    indirect=True,
)
def test_execute_bakes_env(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    _patched_testinfra_get_tests,  # noqa: PT019
    inventory_directory,
    _instance,  # noqa: PT019
):
    _instance.execute()

    assert patched_run_command.call_args[1]["env"]["FOO"] == "bar"


def test_testinfra_executes_catches_and_exits_return_code(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    patched_run_command,
    _patched_testinfra_get_tests,  # noqa: PT019
    _instance,  # noqa: PT019
):
    patched_run_command.side_effect = SystemExit(1)
    with pytest.raises(SystemExit) as e:
        _instance.execute()

    assert e.value.code == 1



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import collections
import os

from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest

from molecule import config, util
from molecule.exceptions import MoleculeError
from molecule.provisioner import ansible, ansible_playbooks
from tests.unit.conftest import os_split  # pylint:disable=C0411


if TYPE_CHECKING:
    from unittest.mock import MagicMock, Mock

    from pytest_mock import MockerFixture

    from molecule.provisioner.ansible import Ansible


@pytest.fixture
def _patched_ansible_playbook(mocker: MockerFixture) -> MagicMock:
    m = mocker.patch("molecule.provisioner.ansible_playbook.AnsiblePlaybook")
    m.return_value.execute.return_value = b"patched-ansible-playbook-stdout"

    return m


@pytest.fixture
def _patched_write_inventory(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("molecule.provisioner.ansible.Ansible._write_inventory")


@pytest.fixture
def _patched_remove_vars(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("molecule.provisioner.ansible.Ansible._remove_vars")


@pytest.fixture
def _patched_link_or_update_vars(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("molecule.provisioner.ansible.Ansible._link_or_update_vars")


@pytest.fixture
def _provisioner_section_data():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "provisioner": {
            "name": "ansible",
            "config_options": {"defaults": {"foo": "bar"}},
            "connection_options": {"foo": "bar"},
            "options": {"foo": "bar", "become": True, "v": True},
            "env": {
                "FOO": "bar",
            },
            "inventory": {
                "hosts": {
                    "all": {
                        "hosts": {"extra-host-01": {}},
                        "children": {"extra-group": {"hosts": ["extra-host-01"]}},
                    },
                },
                "host_vars": {
                    "instance-1": [{"foo": "bar"}],
                    "localhost": [{"foo": "baz"}],
                },
                "group_vars": {
                    "example_group1": [{"foo": "bar"}],
                    "example_group2": [{"foo": "bar"}],
                },
            },
        },
    }


@pytest.fixture(name="instance")
def fixture_instance(
    _provisioner_section_data: dict[str, Any],
    config_instance: config.Config,
) -> ansible.Ansible:
    """Create a provisioner instance.

    Args:
        _provisioner_section_data: A dictionary containing the provisioner section data.
        config_instance: An instance of a Molecule config.
    """
    return ansible.Ansible(config_instance)


def test_profisioner_config_private_member(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert isinstance(instance._config, config.Config)


def test_default_config_options_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = {
        "defaults": {
            "ansible_managed": "Ansible managed: Do NOT edit this file manually!",
            "display_failed_stderr": True,
            "forks": 50,
            "host_key_checking": False,
            # https://docs.ansible.com/ansible/devel/reference_appendices/interpreter_discovery.html
            "interpreter_python": "auto_silent",
            "nocows": 1,
            "retry_files_enabled": False,
        },
        "ssh_connection": {
            "control_path": "%(directory)s/%%h-%%p-%%r",
            "scp_if_ssh": True,
        },
    }

    assert x == instance.default_config_options


def test_provisioner_default_options_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert instance.default_options == {"skip-tags": "molecule-notest,notest"}


def test_ansible_default_env_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = instance._config.provisioner.config_file

    assert x == instance.default_env["ANSIBLE_CONFIG"]
    assert "MOLECULE_FILE" in instance.default_env
    assert "MOLECULE_INVENTORY_FILE" in instance.default_env
    assert "MOLECULE_SCENARIO_DIRECTORY" in instance.default_env
    assert "MOLECULE_INSTANCE_CONFIG" in instance.default_env
    assert "ANSIBLE_CONFIG" in instance.env


def test_provisioner_name_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert instance.name == "ansible"


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_config_options_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = {
        "defaults": {
            "ansible_managed": "Ansible managed: Do NOT edit this file manually!",
            "display_failed_stderr": True,
            "foo": "bar",
            "forks": 50,
            "host_key_checking": False,
            "interpreter_python": "auto_silent",
            "nocows": 1,
            "retry_files_enabled": False,
        },
        "ssh_connection": {
            "control_path": "%(directory)s/%%h-%%p-%%r",
            "scp_if_ssh": True,
        },
    }

    assert x == instance.config_options


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_ansible_options_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = {"become": True, "foo": "bar", "v": True, "skip-tags": "molecule-notest,notest"}

    assert x == instance.options


def test_ansible_options_property_does_not_merge(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    for action in ["create", "destroy"]:
        instance._config.action = action

        assert instance.options == {"skip-tags": "molecule-notest,notest"}


def test_provisioner_ansible_options_property_handles_cli_args(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    instance._config.args = {"debug": True}
    x = {
        "vvv": True,
        "become": True,
        "diff": True,
        "skip-tags": "molecule-notest,notest",
    }

    assert x == instance.options


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_provisioner_env_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = instance._config.provisioner.config_file

    assert x == instance.env["ANSIBLE_CONFIG"]
    assert instance.env["FOO"] == "bar"


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_host_vars_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = {"instance-1": [{"foo": "bar"}], "localhost": [{"foo": "baz"}]}

    assert x == instance.host_vars


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_group_vars_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = {"example_group1": [{"foo": "bar"}], "example_group2": [{"foo": "bar"}]}

    assert x == instance.group_vars


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_hosts_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    hosts = {
        "all": {
            "hosts": {"extra-host-01": {}},
            "children": {"extra-group": {"hosts": ["extra-host-01"]}},
        },
    }

    assert hosts == instance.hosts


def test_links_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert instance.links == {}


def test_inventory_directory_property(instance: Ansible) -> None:
    """Test the inventory_directory property.

    Args:
        instance: Ansible provisioner instance.
    """
    x = Path(instance._config.scenario.ephemeral_directory, "inventory")
    assert str(x) == instance.inventory_directory


def test_inventory_directory_property_shared(instance: Ansible) -> None:
    """Test the shared_inventory_directory property.

    Args:
        instance: Ansible provisioner instance.
    """
    instance._config.command_args["shared_inventory"] = True
    x = Path(instance._config.scenario.shared_ephemeral_directory, "inventory")
    assert str(x) == instance.inventory_directory


def test_inventory_directory_property_shared_parallel(instance: Ansible) -> None:
    """Test the shared_inventory_directory property with parallel mode on.

    Args:
        instance: Ansible provisioner instance.
    """
    instance._config.command_args["shared_inventory"] = True
    instance._config.command_args["parallel"] = True
    # Parallel disables shared ephemeral directory
    x = Path(instance._config.scenario.ephemeral_directory, "inventory")
    assert str(x) == instance.inventory_directory


def test_inventory_file_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = os.path.join(  # noqa: PTH118
        instance._config.scenario.inventory_directory,
        "ansible_inventory.yml",
    )

    assert x == instance.inventory_file


def test_config_file_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = os.path.join(instance._config.scenario.ephemeral_directory, "ansible.cfg")  # noqa: PTH118

    assert x == instance.config_file


def test_playbooks_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert isinstance(instance.playbooks, ansible_playbooks.AnsiblePlaybooks)


def test_provisioner_directory_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    result = instance.directory
    parts = os_split(result)

    assert parts[-3:] == ("molecule", "provisioner", "ansible")


def test_playbooks_cleaned_property_is_optional(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert instance.playbooks.cleanup is None


def test_playbooks_converge_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    x = os.path.join(instance._config.scenario.directory, "converge.yml")  # noqa: PTH118

    assert x == instance.playbooks.converge


def test_playbooks_side_effect_property(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    assert instance.playbooks.side_effect is None


def test_check(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    instance.check()

    assert instance._config.provisioner

    _patched_ansible_playbook.assert_called_once_with(
        instance._config.provisioner.playbooks.converge,
        instance._config,
        verify=False,
    )
    _patched_ansible_playbook.return_value.add_cli_arg.assert_called_once_with(
        "check",
        value=True,
    )
    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_converge(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    result = instance.converge()

    assert instance._config.provisioner

    _patched_ansible_playbook.assert_called_once_with(
        instance._config.provisioner.playbooks.converge,
        instance._config,
        verify=False,
    )
    # NOTE(retr0h): This is not the true return type.  This is a mock return
    #               which didn't go through str.decode().
    assert result == b"patched-ansible-playbook-stdout"  # type: ignore[comparison-overlap]

    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_converge_with_playbook(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    result = instance.converge("playbook")

    _patched_ansible_playbook.assert_called_once_with(
        "playbook",
        instance._config,
        verify=False,
    )
    # NOTE(retr0h): This is not the true return type.  This is a mock return
    #               which didn't go through str.decode().
    assert result == b"patched-ansible-playbook-stdout"  # type: ignore[comparison-overlap]

    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_cleanup(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    instance.cleanup()

    assert instance._config.provisioner

    _patched_ansible_playbook.assert_called_once_with(
        instance._config.provisioner.playbooks.cleanup,
        instance._config,
        verify=False,
    )
    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_destroy(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    instance.destroy()

    assert instance._config.provisioner

    _patched_ansible_playbook.assert_called_once_with(
        instance._config.provisioner.playbooks.destroy,
        instance._config,
        verify=False,
    )
    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_side_effect(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    instance.side_effect()

    assert instance._config.provisioner

    _patched_ansible_playbook.assert_called_once_with(
        instance._config.provisioner.playbooks.side_effect,
        instance._config,
        verify=False,
    )
    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_create(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    instance.create()

    assert instance._config.provisioner

    _patched_ansible_playbook.assert_called_once_with(
        instance._config.provisioner.playbooks.create,
        instance._config,
        verify=False,
    )
    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_prepare(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    instance.prepare()

    assert instance._config.provisioner

    _patched_ansible_playbook.assert_called_once_with(
        instance._config.provisioner.playbooks.prepare,
        instance._config,
        verify=False,
    )
    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_syntax(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    instance.syntax()

    assert instance._config.provisioner

    _patched_ansible_playbook.assert_called_once_with(
        instance._config.provisioner.playbooks.converge,
        instance._config,
        verify=False,
    )
    _patched_ansible_playbook.return_value.add_cli_arg.assert_called_once_with(
        "syntax-check",
        value=True,
    )
    _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_verify(  # noqa: D103
    instance: ansible.Ansible,
    mocker: MockerFixture,
    _patched_ansible_playbook: Mock,  # noqa: PT019
) -> None:
    instance.verify()

    assert instance._config.provisioner

    if instance._config.provisioner.playbooks.verify:
        _patched_ansible_playbook.assert_called_once_with(
            instance._config.provisioner.playbooks.verify,
            instance._config,
        )
        _patched_ansible_playbook.return_value.execute.assert_called_once_with()


def test_ansible_write_config(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    instance.write_config()

    assert os.path.isfile(instance.config_file)  # noqa: PTH113


def test_manage_inventory(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    instance,
    _patched_write_inventory,  # noqa: PT019
    _patched_remove_vars,  # noqa: PT019
    patched_add_or_update_vars,
    _patched_link_or_update_vars,  # noqa: PT019
):
    instance.manage_inventory()

    _patched_write_inventory.assert_called_once_with()
    _patched_remove_vars.assert_called_once_with()
    patched_add_or_update_vars.assert_called_once_with()
    assert not _patched_link_or_update_vars.called


def test_manage_inventory_with_links(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    instance,
    _patched_write_inventory,  # noqa: PT019
    _patched_remove_vars,  # noqa: PT019
    patched_add_or_update_vars,
    _patched_link_or_update_vars,  # noqa: PT019
):
    c = instance._config.config
    c["provisioner"]["inventory"]["links"] = {"foo": "bar"}
    instance.manage_inventory()

    _patched_write_inventory.assert_called_once_with()
    _patched_remove_vars.assert_called_once_with()
    assert not patched_add_or_update_vars.called
    _patched_link_or_update_vars.assert_called_once_with()


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_add_or_update_vars(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    inventory_dir = instance._config.scenario.inventory_directory

    host_vars_directory = os.path.join(inventory_dir, "host_vars")  # noqa: PTH118
    host_vars = os.path.join(host_vars_directory, "instance-1")  # noqa: PTH118

    instance._add_or_update_vars()

    assert os.path.isdir(host_vars_directory)  # noqa: PTH112
    assert os.path.isfile(host_vars)  # noqa: PTH113

    host_vars_localhost = os.path.join(host_vars_directory, "localhost")  # noqa: PTH118
    assert os.path.isfile(host_vars_localhost)  # noqa: PTH113

    group_vars_directory = os.path.join(inventory_dir, "group_vars")  # noqa: PTH118
    group_vars_1 = os.path.join(group_vars_directory, "example_group1")  # noqa: PTH118
    group_vars_2 = os.path.join(group_vars_directory, "example_group2")  # noqa: PTH118

    assert os.path.isdir(group_vars_directory)  # noqa: PTH112
    assert os.path.isfile(group_vars_1)  # noqa: PTH113
    assert os.path.isfile(group_vars_2)  # noqa: PTH113

    hosts = os.path.join(inventory_dir, "hosts")  # noqa: PTH118
    assert os.path.isfile(hosts)  # noqa: PTH113
    assert util.safe_load_file(hosts) == instance.hosts


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_add_or_update_vars_without_host_vars(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    c = instance._config.config
    c["provisioner"]["inventory"]["host_vars"] = {}
    inventory_dir = instance._config.scenario.inventory_directory

    host_vars_directory = os.path.join(inventory_dir, "host_vars")  # noqa: PTH118
    host_vars = os.path.join(host_vars_directory, "instance-1")  # noqa: PTH118

    instance._add_or_update_vars()

    assert not os.path.isdir(host_vars_directory)  # noqa: PTH112
    assert not os.path.isfile(host_vars)  # noqa: PTH113

    host_vars_localhost = os.path.join(host_vars_directory, "localhost")  # noqa: PTH118
    assert not os.path.isfile(host_vars_localhost)  # noqa: PTH113

    group_vars_directory = os.path.join(inventory_dir, "group_vars")  # noqa: PTH118
    group_vars_1 = os.path.join(group_vars_directory, "example_group1")  # noqa: PTH118
    group_vars_2 = os.path.join(group_vars_directory, "example_group2")  # noqa: PTH118

    assert os.path.isdir(group_vars_directory)  # noqa: PTH112
    assert os.path.isfile(group_vars_1)  # noqa: PTH113
    assert os.path.isfile(group_vars_2)  # noqa: PTH113

    hosts = os.path.join(inventory_dir, "hosts")  # noqa: PTH118
    assert os.path.isfile(hosts)  # noqa: PTH113
    assert util.safe_load_file(hosts) == instance.hosts


def test_add_or_update_vars_does_not_create_vars(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    c = instance._config.config
    c["provisioner"]["inventory"]["hosts"] = {}
    c["provisioner"]["inventory"]["host_vars"] = {}
    c["provisioner"]["inventory"]["group_vars"] = {}
    inventory_dir = instance._config.scenario.inventory_directory

    hosts = os.path.join(inventory_dir, "hosts")  # noqa: PTH118
    host_vars_directory = os.path.join(inventory_dir, "host_vars")  # noqa: PTH118
    group_vars_directory = os.path.join(inventory_dir, "group_vars")  # noqa: PTH118

    instance._add_or_update_vars()

    assert not os.path.isdir(host_vars_directory)  # noqa: PTH112
    assert not os.path.isdir(group_vars_directory)  # noqa: PTH112
    assert not os.path.isfile(hosts)  # noqa: PTH113


@pytest.mark.parametrize(
    "config_instance",
    ["_provisioner_section_data"],  # noqa: PT007
    indirect=True,
)
def test_remove_vars(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    inventory_dir = instance._config.scenario.inventory_directory

    hosts = os.path.join(inventory_dir, "hosts")  # noqa: PTH118
    host_vars_directory = os.path.join(inventory_dir, "host_vars")  # noqa: PTH118
    host_vars = os.path.join(host_vars_directory, "instance-1")  # noqa: PTH118

    instance._add_or_update_vars()
    assert os.path.isfile(hosts)  # noqa: PTH113
    assert os.path.isdir(host_vars_directory)  # noqa: PTH112
    assert os.path.isfile(host_vars)  # noqa: PTH113

    host_vars_localhost = os.path.join(host_vars_directory, "localhost")  # noqa: PTH118
    assert os.path.isfile(host_vars_localhost)  # noqa: PTH113

    group_vars_directory = os.path.join(inventory_dir, "group_vars")  # noqa: PTH118
    group_vars_1 = os.path.join(group_vars_directory, "example_group1")  # noqa: PTH118
    group_vars_2 = os.path.join(group_vars_directory, "example_group2")  # noqa: PTH118

    assert os.path.isdir(group_vars_directory)  # noqa: PTH112
    assert os.path.isfile(group_vars_1)  # noqa: PTH113
    assert os.path.isfile(group_vars_2)  # noqa: PTH113

    instance._remove_vars()

    assert not os.path.isfile(hosts)  # noqa: PTH113
    assert not os.path.isdir(host_vars_directory)  # noqa: PTH112
    assert not os.path.isdir(group_vars_directory)  # noqa: PTH112


def test_remove_vars_symlinks(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    inventory_dir = instance._config.scenario.inventory_directory

    source_group_vars = os.path.join(inventory_dir, os.path.pardir, "group_vars")  # noqa: PTH118
    target_group_vars = os.path.join(inventory_dir, "group_vars")  # noqa: PTH118
    os.mkdir(source_group_vars)  # noqa: PTH102
    os.symlink(source_group_vars, target_group_vars)

    instance._remove_vars()

    assert not os.path.lexists(target_group_vars)


def test_link_vars(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    c = instance._config.config
    c["provisioner"]["inventory"]["links"] = {
        "hosts": "../hosts",
        "group_vars": "../group_vars",
        "host_vars": "../host_vars",
    }
    inventory_dir = instance._config.scenario.inventory_directory
    scenario_dir = instance._config.scenario.directory
    source_hosts = os.path.join(scenario_dir, os.path.pardir, "hosts")  # noqa: PTH118
    target_hosts = os.path.join(inventory_dir, "hosts")  # noqa: PTH118
    source_group_vars = os.path.join(scenario_dir, os.path.pardir, "group_vars")  # noqa: PTH118
    target_group_vars = os.path.join(inventory_dir, "group_vars")  # noqa: PTH118
    source_host_vars = os.path.join(scenario_dir, os.path.pardir, "host_vars")  # noqa: PTH118
    target_host_vars = os.path.join(inventory_dir, "host_vars")  # noqa: PTH118

    open(source_hosts, "w").close()  # pylint: disable=consider-using-with  # noqa: PTH123
    os.mkdir(source_group_vars)  # noqa: PTH102
    os.mkdir(source_host_vars)  # noqa: PTH102

    instance._link_or_update_vars()

    assert os.path.lexists(target_hosts)
    assert os.path.lexists(target_group_vars)
    assert os.path.lexists(target_host_vars)


def test_link_vars_raises_when_source_not_found(instance: ansible.Ansible, caplog):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    c = instance._config.config
    c["provisioner"]["inventory"]["links"] = {"foo": "../bar"}

    with pytest.raises(MoleculeError) as e:
        instance._link_or_update_vars()

    assert e.value.code == 1

    source = os.path.join(  # noqa: PTH118
        instance._config.scenario.directory,
        os.path.pardir,
        "bar",
    )
    msg = f"The source path '{source}' does not exist."
    assert msg in caplog.text


def test_verify_inventory(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    instance._verify_inventory()


def test_verify_inventory_raises_when_missing_hosts(  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    caplog,
    instance,
):
    instance._config.config["platforms"] = []
    with pytest.raises(MoleculeError) as e:
        instance._verify_inventory()

    assert e.value.code == 1

    msg = "Instances missing from the 'platform' section of molecule.yml."
    assert msg in caplog.text


def test_vivify(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    d = instance._vivify()
    d["bar"]["baz"] = "qux"

    assert str(d["bar"]["baz"]) == "qux"


def test_default_to_regular(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    d = collections.defaultdict()  # type: ignore[var-annotated]
    assert isinstance(d, collections.defaultdict)

    d = instance._default_to_regular(d)
    assert isinstance(d, dict)


def test_get_plugin_directory(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    result = instance._get_plugin_directory()
    parts = os_split(result)

    assert parts[-4:] == ("molecule", "provisioner", "ansible", "plugins")


def test_absolute_path_for(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    env = {"foo": "foo:bar"}
    x = ":".join(
        [
            os.path.join(instance._config.scenario.directory, "foo"),  # noqa: PTH118
            os.path.join(instance._config.scenario.directory, "bar"),  # noqa: PTH118
        ],
    )

    assert x == instance._absolute_path_for(env, "foo")


def test_absolute_path_for_raises_with_missing_key(instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, D103
    env = {"foo": "foo:bar"}

    with pytest.raises(KeyError):
        instance._absolute_path_for(env, "invalid")



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import pytest

from molecule.status import Status


@pytest.fixture
def _instance() -> Status:
    s = Status(
        instance_name="instance",
        driver_name="driver",
        provisioner_name="provisioner",
        scenario_name="scenario",
        created="True",
        converged="False",
    )

    return s  # noqa: RET504


def test__instance_name_attribute(_instance: Status) -> None:  # noqa: PT019, D103
    assert _instance.instance_name == "instance"


def test_status_driver_name_attribute(_instance: Status) -> None:  # noqa: PT019, D103
    assert _instance.driver_name == "driver"


def test_status_provisioner_name_attribute(_instance: Status) -> None:  # noqa: PT019, D103
    assert _instance.provisioner_name == "provisioner"


def test_status_scenario_name_attribute(_instance: Status) -> None:  # noqa: PT019, D103
    assert _instance.scenario_name == "scenario"


def test_status_created_attribute(_instance: Status) -> None:  # noqa: PT019, D103
    assert _instance.created == "True"


def test_status_converged_attribute(_instance: Status) -> None:  # noqa: PT019, D103
    assert _instance.converged == "False"



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Scenarios Module."""

from __future__ import annotations

import logging

from typing import TYPE_CHECKING

from molecule import util
from molecule.exceptions import MoleculeError


if TYPE_CHECKING:
    from molecule.config import Config
    from molecule.scenario import Scenario
    from molecule.types import ScenariosResults


LOG = logging.getLogger(__name__)


class Scenarios:
    """The Scenarios groups one or more scenario objects Molecule will execute."""

    def __init__(
        self,
        configs: list[Config],
        scenario_names: list[str] | None = None,
    ) -> None:
        """Initialize a new scenarios class and returns None.

        Args:
            configs: Molecule config instances.
            scenario_names: The names of the scenarios.
        """
        self._configs = configs
        self._scenario_names = [] if scenario_names is None else scenario_names
        self._scenarios = self.all
        self.results: list[ScenariosResults] = []

    def __iter__(self) -> Scenarios:
        """Make object iterable.

        Returns:
            The object itself, as Scenarios is iterable.
        """
        return self

    def __next__(self) -> Scenario:
        """Iterate over Scenario objects.

        Returns:
            The next Scenario in the sequence.

        Raises:
            StopIteration: When the scenarios are exhausted.
        """
        if not self._scenarios:
            raise StopIteration
        return self._scenarios.pop(0)

    @property
    def all(self) -> list[Scenario]:
        """Return a list containing all scenario objects.

        Returns:
            All Scenario objects.
        """
        if self._scenario_names:
            scenarios = self._filter_for_scenario()
            self._verify()

            return scenarios

        scenarios = [c.scenario for c in self._configs]
        scenarios.sort(key=lambda x: x.directory)
        return scenarios

    def print_matrix(self) -> None:
        """Show the test matrix for all scenarios."""
        msg = "Test matrix"
        LOG.info(msg)

        tree = {}
        for scenario in self.all:
            tree[scenario.name] = list(scenario.sequence)
        util.print_as_yaml(tree)

    def sequence(self, scenario_name: str) -> list[str]:
        """Sequence for a given scenario.

        Args:
            scenario_name: Name of the scenario to determine the sequence of.

        Returns:
            A list of steps for that scenario.

        Raises:
            RuntimeError: If the scenario cannot be found.
        """
        for scenario in self.all:
            if scenario.name == scenario_name:
                return list(scenario.sequence)
        raise RuntimeError(  # noqa: TRY003
            f"Unable to find sequence for {scenario_name} scenario.",  # noqa: EM102
        )

    def _verify(self) -> None:
        """Verify the specified scenario was found.

        Raises:
            MoleculeError: when scenario is not found.
        """
        scenario_names = [c.scenario.name for c in self._configs]
        if missing_names := sorted(set(self._scenario_names).difference(scenario_names)):
            scenario = "Scenario"
            if len(missing_names) > 1:
                scenario += "s"
            missing = ", ".join(missing_names)
            msg = f"{scenario} '{missing}' not found.  Exiting."
            raise MoleculeError(msg)

    def _filter_for_scenario(self) -> list[Scenario]:
        """Find the scenario matching the provided scenario name and returns a list.

        Returns:
            list
        """
        return [c.scenario for c in self._configs if c.scenario.name in self._scenario_names]

    def _get_matrix(self) -> dict[str, dict[str, list[str]]]:
        """Build a matrix of scenarios and step sequences.

        Returns:
            A dictionary for each scenario listing action sequences for each step.
        """
        return {
            scenario.name: {
                "check": scenario.check_sequence,
                "cleanup": scenario.cleanup_sequence,
                "converge": scenario.converge_sequence,
                "create": scenario.create_sequence,
                "dependency": scenario.dependency_sequence,
                "destroy": scenario.destroy_sequence,
                "idempotence": scenario.idempotence_sequence,
                "prepare": scenario.prepare_sequence,
                "side_effect": scenario.side_effect_sequence,
                "syntax": scenario.syntax_sequence,
                "test": scenario.test_sequence,
                "verify": scenario.verify_sequence,
            }
            for scenario in self.all
        }



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Config Module."""

from __future__ import annotations

import copy
import os
import warnings

from pathlib import Path
from typing import TYPE_CHECKING
from uuid import uuid4

from ansible_compat.ports import cache, cached_property

from molecule import api, interpolation, logger, platforms, scenario, state, util
from molecule.app import get_app
from molecule.data import __file__ as data_module
from molecule.dependency import ansible_galaxy, shell
from molecule.exceptions import MoleculeError
from molecule.model import schema_v3
from molecule.provisioner import ansible
from molecule.util import boolean


if TYPE_CHECKING:
    from collections.abc import MutableMapping
    from typing import Literal

    from packaging.version import Version

    from molecule.dependency.base import Base as Dependency
    from molecule.driver.base import Driver
    from molecule.state import State
    from molecule.types import CollectionData, CommandArgs, ConfigData, MoleculeArgs
    from molecule.verifier.base import Verifier


MOLECULE_PARALLEL: bool = boolean(os.environ.get("MOLECULE_PARALLEL", ""))
MOLECULE_DEBUG: bool = boolean(os.environ.get("MOLECULE_DEBUG", "False"))
MOLECULE_VERBOSITY: int = int(os.environ.get("MOLECULE_VERBOSITY", "0"))
MOLECULE_DIRECTORY = "molecule"
MOLECULE_FILE = "molecule.yml"
MOLECULE_KEEP_STRING = "MOLECULE_"
DEFAULT_DRIVER = "default"

MOLECULE_EMBEDDED_DATA_DIR = os.path.dirname(data_module)  # noqa: PTH120


@cache
def ansible_version() -> Version:
    """Retrieve Ansible version.

    Returns:
        Molecule version information.
    """
    warnings.warn(  # noqa: B028
        "molecule.config.ansible_version is deprecated, will be removed in the future.",
        category=DeprecationWarning,
    )
    return get_app(Path()).runtime.version


class Config:
    """Config Class.

    Molecule searches the current directory for ``molecule.yml`` files by
    globbing `molecule/*/molecule.yml`.  The files are instantiated into
    a list of Molecule [molecule.config.Config][] objects, and each Molecule subcommand
    operates on this list.

    The directory in which the ``molecule.yml`` resides is the Scenario's
    directory.  Molecule performs most functions within this directory.

    The [molecule.config.Config][] object instantiates Dependency, Driver,
    Platforms, Provisioner, Verifier_,
    [scenario][], and State_ references.
    """

    # pylint: disable=too-many-instance-attributes
    # Config objects should be allowed to have any number of attributes
    def __init__(
        self,
        molecule_file: str,  # pylint: disable=redefined-outer-name
        args: MoleculeArgs = {},  # noqa: B006
        command_args: CommandArgs = {},  # noqa: B006
        ansible_args: tuple[str, ...] = (),
    ) -> None:
        """Initialize a new config class and returns None.

        Args:
            molecule_file: A string containing the path to the Molecule file to be parsed.
            args: An optional dict of options, arguments and commands from the CLI.
            command_args: An optional dict of options passed to the subcommand from the CLI.
            ansible_args: An optional tuple of arguments provided to the `ansible-playbook` command.
        """
        self.molecule_file = molecule_file
        self.args = args
        self.command_args = command_args
        self.ansible_args = ansible_args
        self.config = self._get_config()
        self._action: str | None = None
        self._run_uuid = str(uuid4())
        self.project_directory = os.getenv(
            "MOLECULE_PROJECT_DIRECTORY",
            os.getcwd(),  # noqa: PTH109
        )
        self.app = get_app(Path(self.project_directory))
        self.runtime = self.app.runtime
        self.scenario_path = Path(molecule_file).parent

        # Former after_init() contents
        self.config = self._reget_config()
        if self.molecule_file:
            self._validate()

    def write(self) -> None:
        """Write config file to filesystem."""
        util.write_file(self.config_file, util.safe_dump(self.config))

    @property
    def config_file(self) -> str:
        """Path to the config file.

        Returns:
            The path to the config file.
        """
        path = Path(self.scenario.ephemeral_directory) / MOLECULE_FILE
        return str(path)

    @property
    def is_parallel(self) -> bool:
        """Is molecule in parallel mode.

        Returns:
            Whether molecule is in parallel mode.
        """
        return self.command_args.get("parallel", False)

    @property
    def shared_inventory(self) -> bool:
        """Should molecule share ephemeral data.

        Returns:
            Whether molecule should share ephemeral data.
        """
        return self.command_args.get("shared_inventory", False)

    @property
    def shared_data(self) -> bool:
        """Should molecule share ephemeral data.

        Returns:
            Whether molecule should share ephemeral data.
        """
        return self.command_args.get("shared_state", False)

    @property
    def platform_name(self) -> str | None:
        """Configured platform.

        Returns:
            The name of the platform plugin specified.
        """
        return self.command_args.get("platform_name", None)

    @property
    def debug(self) -> bool:
        """Is molecule in debug mode.

        Returns:
            Whether molecule is in debug mode.
        """
        return self.args.get("debug", MOLECULE_DEBUG)

    @property
    def env_file(self) -> str | None:
        """Return path to env file.

        Returns:
            Path to configured env file.
        """
        return util.abs_path(self.args.get("env_file"))

    @property
    def subcommand(self) -> str:
        """Subcommand in use.

        Returns:
            The current subcommand being run.
        """
        return self.command_args["subcommand"]

    @property
    def action(self) -> str | None:
        """Action value.

        Returns:
            The value of action.
        """
        return self._action

    @action.setter
    def action(self, value: str) -> None:
        """Action setter.

        Args:
            value: New value for action.
        """
        self._action = value

    @property
    def cache_directory(
        self,
    ) -> Literal["molecule", "molecule_parallel"]:
        """Proper cache directory to use.

        Returns:
            Either "molecule" or "molecule_parallel" if in parallel mode.
        """
        return "molecule_parallel" if self.is_parallel else "molecule"

    @property
    def collection_directory(self) -> Path | None:
        """Location of collection containing the molecule files.

        Returns:
            Root of the collection containing the molecule files.
        """
        test_paths = [Path.cwd(), Path(self.project_directory)]

        for path in test_paths:
            if (path / "galaxy.yml").exists():
                return path

        # Last resort, try to find git root
        show_toplevel = self.app.run_command("git rev-parse --show-toplevel")
        if show_toplevel.returncode == 0:
            path = Path(show_toplevel.stdout.strip())
            if (path / "galaxy.yml").exists():
                return path

        return None

    @property
    def molecule_directory(self) -> str:
        """Molecule directory for this project.

        Returns:
            The appropriate molecule directory for this project.
        """
        return molecule_directory(self.project_directory)

    @cached_property
    def collection(self) -> CollectionData | None:
        """Collection metadata sourced from galaxy.yml.

        Returns:
            A dictionary of information about the collection molecule is running inside, if any.
        """
        collection_directory = self.collection_directory
        if not collection_directory:
            return None

        galaxy_file = collection_directory / "galaxy.yml"
        galaxy_data: CollectionData = util.safe_load_file(galaxy_file)

        important_keys = {"name", "namespace"}
        if missing_keys := important_keys.difference(galaxy_data.keys()):
            self._log.warning(
                "The detected galaxy.yml file (%s) is invalid, missing mandatory field %s",
                galaxy_file,
                util.oxford_comma(missing_keys),
            )
            return None  # pragma: no cover

        return galaxy_data

    @cached_property
    def dependency(self) -> Dependency | None:
        """Dependency manager in use.

        Returns:
            Instance of a molecule dependency plugin.
        """
        dependency_name = self.config["dependency"]["name"]
        if dependency_name == "galaxy":
            return ansible_galaxy.AnsibleGalaxy(self)
        if dependency_name == "shell":
            return shell.Shell(self)
        return None

    @cached_property
    def driver(self) -> Driver:
        """Return driver.

        Returns:
            The driver for this scenario.

        Raises:
            MoleculeError: when the specified driver cannot be found.
        """
        driver_name = self._get_driver_name()
        driver = None

        api_drivers = api.drivers(config=self)
        if driver_name not in api_drivers:
            msg = f"Failed to find driver {driver_name}. Please ensure that the driver is correctly installed."
            raise MoleculeError(msg)

        driver = api_drivers[driver_name]
        driver.name = driver_name

        return driver

    @property
    def executor(self) -> str:
        """Return playbook executor.

        Returns:
            The executor backend.
        """
        return self.config.get("executor", {}).get("backend", "ansible-playbook")

    @property
    def env(self) -> dict[str, str]:
        """Environment variables.

        Returns:
            Total set of computed environment variables.
        """
        return {
            "MOLECULE_DEBUG": str(self.debug),
            "MOLECULE_FILE": self.config_file,
            "MOLECULE_ENV_FILE": str(self.env_file),
            "MOLECULE_STATE_FILE": self.state.state_file,
            "MOLECULE_INVENTORY_FILE": self.provisioner.inventory_file,  # type: ignore[union-attr]
            "MOLECULE_EPHEMERAL_DIRECTORY": self.scenario.ephemeral_directory,
            "MOLECULE_SCENARIO_DIRECTORY": self.scenario.directory,
            "MOLECULE_PROJECT_DIRECTORY": self.project_directory,
            "MOLECULE_INSTANCE_CONFIG": self.driver.instance_config,
            "MOLECULE_DEPENDENCY_NAME": self.dependency.name,  # type: ignore[union-attr]
            "MOLECULE_DRIVER_NAME": self.driver.name,
            "MOLECULE_PROVISIONER_NAME": self.provisioner.name,  # type: ignore[union-attr]
            "MOLECULE_SCENARIO_NAME": self.scenario.name,
            "MOLECULE_VERIFIER_NAME": self.verifier.name,
            "MOLECULE_VERIFIER_TEST_DIRECTORY": self.verifier.directory,
        }

    @cached_property
    def platforms(self) -> platforms.Platforms:
        """Platforms for this run.

        Returns:
            A molecule Platforms instance.
        """
        return platforms.Platforms(
            self,
            parallelize_platforms=self.is_parallel,
            platform_name=self.platform_name,
        )

    @cached_property
    def provisioner(self) -> ansible.Ansible | None:
        """Provisioner for this run.

        Returns:
            An instance of the Ansible provisioner.
        """
        provisioner_name = self.config["provisioner"]["name"]
        if provisioner_name == "ansible":
            return ansible.Ansible(self)
        return None

    @cached_property
    def scenario(self) -> scenario.Scenario:
        """Scenario for this run.

        Returns:
            A molecule Scenario instance.
        """
        return scenario.Scenario(self)

    @property
    def _log(self) -> logger.ScenarioLoggerAdapter:
        """Get a scenario logger with config step context.

        Returns:
            A scenario logger adapter with current scenario and step context.
        """
        return logger.get_scenario_logger(__name__, self.scenario.name, "config")

    @cached_property
    def state(self) -> State:
        """Molecule state object.

        Returns:
            A molecule State instance.
        """
        my_state = state.State(self)

        # look at state file for molecule.yml date modified and warn if they do not match
        if self.molecule_file and os.path.isfile(self.molecule_file):  # noqa: PTH113
            modTime = os.path.getmtime(self.molecule_file)  # noqa: PTH204, N806
            if my_state.molecule_yml_date_modified is None:
                my_state.change_state("molecule_yml_date_modified", modTime)
            elif my_state.molecule_yml_date_modified != modTime:
                self._log.warning(
                    "The scenario config file ('%s') has been modified since the scenario was created. "
                    "If recent changes are important, reset the scenario with 'molecule destroy' to clean up created items or "
                    "'molecule reset' to clear current configuration.",
                    self.molecule_file,
                )

        return state.SharedState(self) if self.shared_data else state.State(self)

    @cached_property
    def verifier(self) -> Verifier:
        """Retrieve current verifier.

        Raises:
            RuntimeError: If is not able to find the driver.

        Returns:
            Instance of Verifier driver.
        """
        name = self.config["verifier"]["name"]
        if name not in api.verifiers(self):
            msg = f"Unable to find '{name}' verifier driver."
            raise RuntimeError(msg)
        return api.verifiers(self)[name]

    def _get_driver_name(self) -> str:
        # the state file contains the driver from the last run
        driver_from_state_file = self.state.driver
        # the user may supply a driver on the command line
        driver_from_cli = self.command_args.get("driver_name")
        # the driver may also be edited in the scenario
        driver_from_scenario = self.config["driver"]["name"]

        if driver_from_state_file:
            driver_name = driver_from_state_file
        elif driver_from_cli:
            driver_name = driver_from_cli
        else:
            driver_name = driver_from_scenario
        if driver_name is None:
            driver_name = "default"

        if driver_from_cli and (driver_from_cli != driver_name):
            msg = (
                f"Instance(s) were created with the '{driver_name}' driver, but the "
                f"subcommand is using '{driver_from_cli}' driver."
            )
            raise MoleculeError(msg)

        if driver_from_state_file and driver_name not in api.drivers():
            msg = (
                f"Driver '{driver_name}' from state-file "
                f"'{self.state.state_file}' is not available."
            )
            raise MoleculeError(msg)

        if driver_from_scenario != driver_name:
            msg = (
                f"Driver '{driver_name}' is currently in use but the scenario config "
                f"has changed and now defines '{driver_from_scenario}'. "
                "To change drivers, run 'molecule destroy' for converged scenarios or 'molecule reset' otherwise."
            )
            self._log.warning(msg)

        return driver_name

    def _get_config(self) -> ConfigData:
        """Perform a prioritized recursive merge of config files.

        Returns a new dict.  Prior to merging the config files are interpolated with
        environment variables.

        Returns:
            dict: The merged config.
        """
        return self._combine(keep_string=MOLECULE_KEEP_STRING)

    def _reget_config(self) -> ConfigData:
        """Perform the same prioritized recursive merge from `get_config`.

        Interpolates the ``keep_string`` left behind in the original
        ``get_config`` call.  This is probably __very__ bad.

        Returns:
            dict: The merged config.
        """
        env = util.merge_dicts(os.environ, self.env)
        env = set_env_from_file(env, self.env_file)

        return self._combine(env=env)

    def _combine(
        self,
        env: MutableMapping[str, str] = os.environ,
        keep_string: str | None = None,
    ) -> ConfigData:
        """Perform a prioritized recursive merge of config files.

        Returns a new dict.  Prior to merging the config files are interpolated with
        environment variables.

        1. Loads Molecule defaults.
        2. Loads a base config (if provided) and merges on top of defaults.
        3. Loads the scenario's ``molecule file`` and merges on top of previous
           merge.

        Args:
            env: The current set of environment variables to consider.
            keep_string: String to avoid templating.

        Returns:
            dict: The merged config.
        """
        defaults = self._get_defaults()
        base_configs = filter(os.path.exists, self.args.get("base_config", []))
        for base_config in base_configs:
            with open(base_config) as stream:  # noqa: PTH123
                s = stream.read()
                interpolated_config = self._interpolate(s, env, keep_string)
                defaults = util.merge_dicts(
                    defaults,
                    util.safe_load(interpolated_config),
                )

        if self.molecule_file:
            with open(self.molecule_file) as stream:  # noqa: PTH123
                s = stream.read()
                interpolated_config = self._interpolate(s, env, keep_string)
                defaults = util.merge_dicts(
                    defaults,
                    util.safe_load(interpolated_config),
                )

        return defaults

    def _interpolate(
        self,
        stream: str,
        env: MutableMapping[str, str],
        keep_string: str | None,
    ) -> str:
        env = set_env_from_file(env, self.env_file)

        i = interpolation.Interpolator(interpolation.TemplateWithDefaults, env)

        try:
            return i.interpolate(stream, keep_string)
        except interpolation.InvalidInterpolation as e:
            msg = f"parsing config file '{self.molecule_file}'.\n\n{e.place}\n{e.string}"
            raise MoleculeError(msg) from e
        return ""

    def _get_defaults(self) -> ConfigData:
        if not self.molecule_file:
            scenario_name = "default"
        else:
            scenario_name = (
                os.path.basename(os.path.dirname(self.molecule_file))  # noqa: PTH119, PTH120
                or "default"
            )
        return {
            "dependency": {
                "name": "galaxy",
                "command": None,
                "enabled": True,
                "options": {},
                "env": {},
            },
            "driver": {
                "name": "default",
                "provider": {"name": None},
                "options": {"managed": True},
                "ssh_connection_options": [],
                "safe_files": [],
            },
            "executor": {
                "backend": "ansible-playbook",
            },
            "platforms": [],
            "prerun": True,
            "role_name_check": 0,
            "provisioner": {
                "name": "ansible",
                "config_options": {},
                "ansible_args": [],
                "connection_options": {},
                "options": {},
                "env": {},
                "inventory": {
                    "hosts": {},
                    "host_vars": {},
                    "group_vars": {},
                    "links": {},
                },
                "children": {},
                "playbooks": {
                    "cleanup": "cleanup.yml",
                    "create": "create.yml",
                    "converge": "converge.yml",
                    "destroy": "destroy.yml",
                    "prepare": "prepare.yml",
                    "side_effect": "side_effect.yml",
                    "verify": "verify.yml",
                },
                "log": True,
            },
            "scenario": {
                "name": scenario_name,
                "check_sequence": [
                    "dependency",
                    "cleanup",
                    "destroy",
                    "create",
                    "prepare",
                    "converge",
                    "check",
                    "cleanup",
                    "destroy",
                ],
                "cleanup_sequence": ["cleanup"],
                "converge_sequence": ["dependency", "create", "prepare", "converge"],
                "create_sequence": ["dependency", "create", "prepare"],
                "destroy_sequence": ["dependency", "cleanup", "destroy"],
                "test_sequence": [
                    # dependency must be kept before lint to avoid errors
                    "dependency",
                    "cleanup",
                    "destroy",
                    "syntax",
                    "create",
                    "prepare",
                    "converge",
                    "idempotence",
                    "side_effect",
                    "verify",
                    "cleanup",
                    "destroy",
                ],
            },
            "verifier": {
                "name": "ansible",
                "enabled": True,
                "options": {},
                "env": {},
                "additional_files_or_dirs": [],
            },
        }

    def _validate(self) -> None:
        """Validate molecule file.

        Raises:
            MoleculeError: when config file fails to validate.
        """
        # Use scenario logger with hardcoded values since scenario property isn't available yet
        scenario_name = self.config["scenario"]["name"]
        validation_log = logger.get_scenario_logger(__name__, scenario_name, "validate")

        msg = f"Validating schema {self.molecule_file}."
        validation_log.debug(msg)

        errors = schema_v3.validate(self.config)
        if errors:
            msg = f"Failed to validate {self.molecule_file}\n\n{errors}"
            raise MoleculeError(msg)


def molecule_directory(path: str | Path) -> str:
    """Return directory of the current scenario.

    Args:
        path: Base molecule directory.

    Returns:
        The current scenario's directory.
    """
    if isinstance(path, str):
        path = Path(path)
    return str(path / MOLECULE_DIRECTORY)


def molecule_file(path: str) -> str:
    """Return file path of current scenario.

    Args:
        path: Base molecule directory.

    Returns:
        The path to the molecule file.
    """
    return os.path.join(path, MOLECULE_FILE)  # noqa: PTH118


def set_env_from_file(
    env: MutableMapping[str, str],
    env_file: str | None,
) -> MutableMapping[str, str]:
    """Load environment from file.

    Args:
        env: Currently known environment variables.
        env_file: File from which to load more environment variables.

    Returns:
        The combined set of environment variables.
    """
    if env_file and os.path.exists(env_file):  # noqa: PTH110
        env = copy.copy(env)
        d = util.safe_load_file(env_file)
        for k, v in d.items():
            env[k] = v

        return env

    return env



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import os
import shutil

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from molecule import config, util
from molecule.scenario import Scenario


if TYPE_CHECKING:
    from unittest.mock import Mock


# NOTE(retr0h): The use of the `patched_config_validate` fixture, disables
# config.Config._validate from executing.  Thus preventing odd side-effects
# throughout patched.assert_called unit tests.
@pytest.fixture
def _instance(
    patched_config_validate: Mock,
    config_instance: config.Config,
) -> Scenario:
    return Scenario(config_instance)


def test_prune(_instance: Scenario) -> None:  # noqa: PT019, D103
    e_dir = Path(_instance.ephemeral_directory)
    # prune data also includes files in the scenario inventory dir,
    # which is "<e_dir>/inventory" by default.
    # items are created in listed order, directories first, safe before pruned
    prune_data = {
        # these files should not be pruned
        "safe_files": ["state.yml", "ansible.cfg", "inventory/ansible_inventory.yml"],
        # these directories should not be pruned
        "safe_dirs": ["inventory"],
        # these files should be pruned
        "pruned_files": ["foo", "bar", "inventory/foo", "inventory/bar"],
        # these directories should be pruned, including empty subdirectories
        "pruned_dirs": ["baz", "roles", "inventory/baz", "roles/foo"],
    }

    for directory in prune_data["safe_dirs"] + prune_data["pruned_dirs"]:
        # inventory dir should already exist, and its existence is
        # required by the assertions below.
        if directory == "inventory":
            continue
        (e_dir / directory).mkdir()

    for file in prune_data["safe_files"] + prune_data["pruned_files"]:
        util.write_file(str(e_dir / file), "")

    _instance.prune()

    for safe_file in prune_data["safe_files"]:
        assert (e_dir / safe_file).is_file()

    for safe_dir in prune_data["safe_dirs"]:
        assert (e_dir / safe_dir).is_dir()

    for pruned_file in prune_data["pruned_files"]:
        assert not (e_dir / pruned_file).is_file()

    for pruned_dir in prune_data["pruned_dirs"]:
        assert not (e_dir / pruned_dir).is_dir()


def test_config_member(_instance: Scenario) -> None:  # noqa: PT019, D103
    assert isinstance(_instance.config, config.Config)


def test_scenario_init_calls_setup(  # noqa: D103
    patched_scenario_setup: Mock,
    _instance: Scenario,  # noqa: PT019
) -> None:
    patched_scenario_setup.assert_called_once_with()


def test_scenario_name_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert _instance.name == "default"


def test_ephemeral_directory_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert os.access(_instance.ephemeral_directory, os.W_OK)
    # assert that scenario path is included in in repr (useful for debugging)
    assert _instance.ephemeral_directory in repr(_instance)


def test_scenario_inventory_directory_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    e_dir = Path(_instance.config.scenario.ephemeral_directory) / "inventory"

    assert str(e_dir) == _instance.inventory_directory


def test_check_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    sequence = [
        "dependency",
        "cleanup",
        "destroy",
        "create",
        "prepare",
        "converge",
        "check",
        "cleanup",
        "destroy",
    ]

    assert sequence == _instance.check_sequence


def test_converge_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    sequence = ["dependency", "create", "prepare", "converge"]

    assert sequence == _instance.converge_sequence


def test_create_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    sequence = ["dependency", "create", "prepare"]

    assert sequence == _instance.create_sequence


def test_dependency_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert _instance.dependency_sequence == ["dependency"]


def test_destroy_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert _instance.destroy_sequence == ["dependency", "cleanup", "destroy"]


def test_idempotence_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert _instance.idempotence_sequence == ["idempotence"]


def test_prepare_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert _instance.prepare_sequence == ["prepare"]


def test_side_effect_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert _instance.side_effect_sequence == ["side_effect"]


def test_syntax_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert _instance.syntax_sequence == ["syntax"]


def test_test_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    sequence = [
        "dependency",
        "cleanup",
        "destroy",
        "syntax",
        "create",
        "prepare",
        "converge",
        "idempotence",
        "side_effect",
        "verify",
        "cleanup",
        "destroy",
    ]

    assert sequence == _instance.test_sequence


def test_verify_sequence_property(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    assert _instance.verify_sequence == ["verify"]


def test_sequence_property_with_invalid_subcommand(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    _instance.config.command_args = {"subcommand": "invalid"}

    assert _instance.sequence == []


def test_setup_creates_ephemeral_and_inventory_directories(  # noqa: D103
    _instance: Scenario,  # noqa: PT019
) -> None:
    ephemeral_dir = _instance.config.scenario.ephemeral_directory
    inventory_dir = _instance.config.scenario.inventory_directory
    shutil.rmtree(ephemeral_dir)
    _instance._setup()

    assert Path(ephemeral_dir).is_dir()
    assert Path(inventory_dir).is_dir()
    # assure we can write to ephemeral directory
    assert os.access(ephemeral_dir, os.W_OK)


def test_ephemeral_directory_overridden_via_env_var(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
) -> None:
    """Confirm ephemeral_directory is overridden by MOLECULE_EPHEMERAL_DIRECTORY.

    Args:
        monkeypatch: Pytest monkeypatch fixture.
        tmp_path: Pytest tmp_path fixture.
    """
    monkeypatch.chdir(tmp_path)
    monkeypatch.setenv("MOLECULE_EPHEMERAL_DIRECTORY", "foo/bar")
    scenario = Scenario(config.Config(""))

    assert os.access(scenario.ephemeral_directory, os.W_OK)
    # Confirm MOLECULE_EPHEMERAL_DIRECTORY uses absolute path.
    assert Path(scenario.ephemeral_directory).is_absolute()
    assert scenario.ephemeral_directory.endswith("foo/bar")



#  Copyright (c) 2015-2018 Cisco Systems, Inc.  # noqa: D100
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import binascii
import os
import tempfile

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from molecule import util
from molecule.console import console
from molecule.constants import MOLECULE_HEADER
from molecule.exceptions import MoleculeError
from molecule.text import strip_ansi_escape


if TYPE_CHECKING:
    from collections.abc import MutableMapping
    from typing import Any
    from unittest.mock import Mock

    from pytest_mock import MockerFixture

    from molecule.app import App
    from molecule.types import Options


def test_print_debug() -> None:  # noqa: D103
    expected = "DEBUG: test_title:\ntest_data\n"
    with console.capture() as capture:
        util.print_debug("test_title", "test_data")

    result = strip_ansi_escape(capture.get())
    assert result == expected


def test_print_environment_vars(  # noqa: D103
    capsys: pytest.CaptureFixture[str],
) -> None:
    env = {
        "ANSIBLE_FOO": "foo",
        "ANSIBLE_BAR": "bar",
        "ANSIBLE": "",
        "MOLECULE_FOO": "foo",
        "MOLECULE_BAR": "bar",
        "MOLECULE": "",
    }
    expected = """DEBUG: ANSIBLE ENVIRONMENT:
ANSIBLE_BAR: bar
ANSIBLE_FOO: foo

DEBUG: MOLECULE ENVIRONMENT:
MOLECULE_BAR: bar
MOLECULE_FOO: foo

DEBUG: SHELL REPLAY:
ANSIBLE_BAR=bar ANSIBLE_FOO=foo MOLECULE_BAR=bar MOLECULE_FOO=foo
"""

    with console.capture() as capture:
        util.print_environment_vars(env)
    result = strip_ansi_escape(capture.get())
    assert result == expected


def test_sysexit() -> None:  # noqa: D103
    with pytest.raises(SystemExit) as e:
        util.sysexit()

    assert e.value.code == 1


def test_sysexit_with_custom_code() -> None:  # noqa: D103
    with pytest.raises(SystemExit) as e:
        util.sysexit(2)

    assert e.value.code == 2  # noqa: PLR2004


def test_run_command(app: App) -> None:  # noqa: D103
    cmd = ["ls"]
    x = app.run_command(cmd)

    assert x.returncode == 0


def test_run_command_with_debug(  # noqa: D103
    mocker: MockerFixture,
    patched_print_debug: Mock,
    app: App,
) -> None:
    env = {"ANSIBLE_FOO": "foo", "MOLECULE_BAR": "bar"}
    app.run_command(["ls"], debug=True, env=env)
    x = [
        mocker.call("ANSIBLE ENVIRONMENT", "ANSIBLE_FOO: foo\n"),
        mocker.call("MOLECULE ENVIRONMENT", "MOLECULE_BAR: bar\n"),
        mocker.call("SHELL REPLAY", "ANSIBLE_FOO=foo MOLECULE_BAR=bar"),
    ]

    assert x == patched_print_debug.mock_calls


def test_run_command_baked_cmd_env(app: App) -> None:  # noqa: D103
    cmd = ["printenv", "myvar"]
    result = app.run_command(cmd, env={"myvar": "value2"})
    assert result.returncode == 0

    cmd = ["printenv", "myvar2"]
    result = app.run_command(cmd, env={"myvar2": "value2"})
    assert result.returncode == 0

    # negative test
    cmd = ["printenv", "myvar"]
    result = app.run_command(cmd)
    assert result.returncode == 1


def test_run_command_with_debug_handles_no_env(  # noqa: D103
    mocker: MockerFixture,
    patched_print_debug: Mock,
    app: App,
) -> None:
    cmd = ["ls"]
    app.run_command(cmd, debug=True)
    # when env is empty we expect not to print anything
    empty_list: list[Any] = []

    assert empty_list == patched_print_debug.mock_calls


def test_os_walk(test_cache_path: Path) -> None:  # noqa: D103
    scenarios = ["scenario1", "scenario2", "scenario3"]
    mol_dir = test_cache_path / "molecule"
    for scenario in scenarios:
        scenario_directory = mol_dir / scenario
        molecule_file = scenario_directory / "molecule.yml"
        os.makedirs(scenario_directory, exist_ok=True)  # noqa: PTH103
        util.write_file(str(molecule_file), "")

    result = list(util.os_walk(mol_dir, "molecule.yml"))
    assert len(result) == 3  # noqa: PLR2004


def test_render_template() -> None:  # noqa: D103
    template = "{{ foo }} = {{ bar }}"

    assert util.render_template(template, foo="foo", bar="bar") == "foo = bar"


def test_render_template_quoted() -> None:  # noqa: D103
    template = """
    {{ 'url = "quoted_str"' }}
    """.strip()

    assert util.render_template(template) == 'url = "quoted_str"'


def test_write_file(test_cache_path: Path) -> None:
    """Test the `write_file` function.

    Args:
        test_cache_path: The path to the test cache directory for the test.
    """
    dest_file = test_cache_path / "test_util_write_file.tmp"
    contents = binascii.b2a_hex(os.urandom(15)).decode("utf-8")
    util.write_file(str(dest_file), contents)
    with dest_file.open() as stream:
        data = stream.read()
    x = f"# Molecule managed\n\n{contents}"
    assert x == data


def test_molecule_prepender(tmp_path: Path) -> None:  # noqa: D103
    fname = tmp_path / "some.txt"
    fname.write_text("foo bar")
    x = f"{MOLECULE_HEADER}\n\nfoo bar"
    util.file_prepender(str(fname))
    assert x == fname.read_text()


def test_safe_dump() -> None:  # noqa: D103
    x = """
---
foo: bar
""".lstrip()

    assert x == util.safe_dump({"foo": "bar"})


def test_safe_dump_with_increase_indent() -> None:  # noqa: D103
    data = {"foo": [{"foo": "bar", "baz": "zzyzx"}]}

    x = """
---
foo:
  - baz: zzyzx
    foo: bar
""".lstrip()
    assert x == util.safe_dump(data)


def test_safe_load() -> None:  # noqa: D103
    assert util.safe_load("foo: bar") == {"foo": "bar"}


def test_safe_load_returns_empty_dict_on_empty_string() -> None:  # noqa: D103
    assert util.safe_load("") == {}


def test_safe_load_exits_when_cannot_parse() -> None:  # noqa: D103
    data = """
---
%foo:
""".strip()

    with pytest.raises(MoleculeError) as e:
        util.safe_load(data)

    assert e.value.code == 1


def test_safe_load_file(test_cache_path: Path) -> None:
    """Test the `safe_load_file` function.

    Args:
        test_cache_path: The path to the test cache directory for the test.
    """
    path = test_cache_path / "test_safe_load_file.yml"
    util.write_file(str(path), "foo: bar")

    assert util.safe_load_file(str(path)) == {"foo": "bar"}


def test_instance_with_scenario_name() -> None:  # noqa: D103
    assert util.instance_with_scenario_name("foo", "bar") == "foo-bar"


def test_verbose_flag() -> None:  # noqa: D103
    options: Options = {"verbose": True, "v": True}

    assert util.verbose_flag(options) == ["-v"]
    # pylint: disable=use-implicit-booleaness-not-comparison
    assert options == {}


def test_verbose_flag_extra_verbose() -> None:  # noqa: D103
    options: Options = {"verbose": True, "vvv": True}

    assert util.verbose_flag(options) == ["-vvv"]
    # pylint: disable=use-implicit-booleaness-not-comparison
    assert options == {}


def test_verbose_flag_preserves_verbose_option() -> None:  # noqa: D103
    options: Options = {"verbose": True}

    # pylint: disable=use-implicit-booleaness-not-comparison
    assert util.verbose_flag(options) == []
    assert options == {"verbose": True}


def test_filter_verbose_permutation() -> None:  # noqa: D103
    options: Options = {
        "v": True,
        "vv": True,
        "vvv": True,
        "vfoo": True,
        "foo": True,
        "bar": True,
    }

    x = {"vfoo": True, "foo": True, "bar": True}
    assert x == util.filter_verbose_permutation(options)


def test_abs_path() -> None:
    """Test the `abs_path` function."""
    test_dir = "/foo/../foo"
    assert util.abs_path(test_dir) == "/foo"


def test_abs_path_with_path() -> None:
    """Test the `abs_path` function."""
    test_dir = Path("/foo/../foo")
    assert util.abs_path(test_dir) == Path("/foo")


def test_abs_path_with_empty_path() -> None:
    """Test the `abs_path` function with an empty path."""
    assert util.abs_path("") == ""


def test_abs_path_with_symlink() -> None:
    """Test the `abs_path` function not resolving symlinks."""
    with tempfile.NamedTemporaryFile() as tmp_file:
        tmpfile_path = Path(tmp_file.name)
        symlink_path = Path(tmp_file.name + "_sym")
        symlink_path.symlink_to(tmpfile_path)
        abs_path_result = util.abs_path(symlink_path)
        symlink_path.unlink()
    assert abs_path_result == symlink_path


@pytest.mark.parametrize(
    ("a", "b", "x"),
    (
        # Base of recursion scenarios
        ({"key": 1}, {"key": 2}, {"key": 2}),
        ({"key": {}}, {"key": 2}, {"key": 2}),
        ({"key": 1}, {"key": {}}, {"key": {}}),
        # Recursive scenario
        ({"a": {"x": 1}}, {"a": {"x": 2}}, {"a": {"x": 2}}),
        ({"a": {"x": 1}}, {"a": {"y": 2}}, {"a": {"x": 1, "y": 2}}),
        # example taken from python-anyconfig/anyconfig/__init__.py
        (
            {"b": [{"c": 0}, {"c": 2}], "d": {"e": "aaa", "f": 3}},
            {"a": 1, "b": [{"c": 3}], "d": {"e": "bbb"}},
            {"a": 1, "b": [{"c": 3}], "d": {"e": "bbb", "f": 3}},
        ),
    ),
)
def test_merge_dicts(a: MutableMapping, b: MutableMapping, x: MutableMapping) -> None:  # type: ignore[type-arg]  # noqa: D103
    assert x == util.merge_dicts(a, b)


@pytest.mark.parametrize(
    ("sequence", "output"),
    (
        ([], ""),
        (["item1"], "'item1'"),
        (["item1", False], "'item1' and 'False'"),
        (["item1", False, Path()], "'item1', 'False', and '.'"),
    ),
    ids=("empty", "one", "two", "three"),
)
def test_oxford_comma(sequence: list[str], output: str) -> None:
    """Test the oxford_comma function.

    Args:
        sequence: sequence of items.
        output: expected output string.
    """
    assert util.oxford_comma(sequence) == output



"""Testinfra tests."""  # noqa: INP001

from __future__ import annotations

import os

import testinfra.utils.ansible_runner


testinfra_hosts = testinfra.utils.ansible_runner.AnsibleRunner(
    os.environ["MOLECULE_INVENTORY_FILE"],
).get_hosts("all")


def test_hostname(host):  # type: ignore[no-untyped-def]  # noqa: ANN201
    """Validate hostname."""
    assert host.check_output("hostname -s") == "instance"


def test_etc_molecule_directory(host):  # type: ignore[no-untyped-def]  # noqa: ANN201
    """Validate molecule directory."""
    f = host.file("/etc/molecule")

    assert f.is_directory
    assert f.user == "root"
    assert f.group == "root"
    assert f.mode == 0o755  # noqa: PLR2004


def test_etc_molecule_ansible_hostname_file(host):  # type: ignore[no-untyped-def]  # noqa: ANN201
    """Validate molecule instance file."""
    f = host.file("/etc/molecule/instance")

    assert f.is_file
    assert f.user == "root"
    assert f.group == "root"
    assert f.mode == 0o644  # noqa: PLR2004



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
"""Molecule Utils Module."""

from __future__ import annotations

import copy
import fnmatch
import logging
import os
import re
import sys

from pathlib import Path
from typing import TYPE_CHECKING, overload

import jinja2
import yaml

from ansible_compat.ports import cache
from rich.syntax import Syntax

from molecule.console import console
from molecule.constants import MOLECULE_HEADER
from molecule.exceptions import MoleculeError


if TYPE_CHECKING:
    from collections.abc import Generator, Iterable, MutableMapping, Sequence
    from io import TextIOWrapper
    from typing import Any, AnyStr, NoReturn, TypeVar
    from warnings import WarningMessage

    from molecule.types import CommandArgs, ConfigData, Options, PlatformData

    NestedDict = MutableMapping[str, Any]
    _T = TypeVar("_T", bound=NestedDict)


LOG = logging.getLogger(__name__)


class SafeDumper(yaml.SafeDumper):
    """SafeDumper YAML Class."""

    def increase_indent(
        self,
        flow: bool = False,  # noqa: FBT001, FBT002
        indentless: bool = False,  # noqa: ARG002, FBT001, FBT002
    ) -> None:
        """Increase indent of the YAML structure.

        Args:
            flow: Whether to reset indenting on newline.
            indentless: Whether to skip additional indenting.
        """
        return super().increase_indent(flow, indentless=False)


def print_debug(title: str, data: str) -> None:
    """Print debug information.

    Args:
        title: A title to describe the data.
        data: The data to print.
    """
    console.print(f"DEBUG: {title}:\n{data}")


def print_environment_vars(env: dict[str, str] | None) -> None:
    """Print ``Ansible`` and ``Molecule`` environment variables and returns None.

    Args:
        env: A dict containing the shell's environment as collected by ``os.environ``.
    """
    if env:
        ansible_env = {k: v for (k, v) in env.items() if "ANSIBLE_" in k}
        print_debug("ANSIBLE ENVIRONMENT", safe_dump(ansible_env, explicit_start=False))

        molecule_env = {k: v for (k, v) in env.items() if "MOLECULE_" in k}
        print_debug(
            "MOLECULE ENVIRONMENT",
            safe_dump(molecule_env, explicit_start=False),
        )

        combined_env = ansible_env.copy()
        combined_env.update(molecule_env)
        print_debug(
            "SHELL REPLAY",
            " ".join([f"{k}={v}" for (k, v) in sorted(combined_env.items())]),
        )


def do_report() -> None:
    """Dump html report atexit."""
    report_file = Path(os.environ["MOLECULE_REPORT"])
    LOG.info("Writing %s report.", report_file)
    with report_file.open("w") as f:
        f.write(console.export_html())
        f.close()


def sysexit(code: int = 1) -> NoReturn:
    """Perform a system exit with given code.

    Args:
        code: The return code to emit.
    """
    sys.exit(code)


def sysexit_with_message(
    msg: str,
    code: int = 1,
    warns: Sequence[WarningMessage] = (),
) -> NoReturn:
    """This method is a lie for compatibility purposes.

    Args:
        msg: The message to display.
        code: The return code to exit with.
        warns: A series of warnings to send alongside the message.

    Raises:
        MoleculeError: always.
    """
    raise MoleculeError(message=msg, code=code, warns=warns)


def os_walk(
    directory: str | Path,
    pattern: str,
    excludes: list[str] | None = None,
    *,
    followlinks: bool = False,
) -> Generator[str, None, None]:
    """Navigate recursively and retried files based on pattern.

    Args:
        directory: The base directory.
        pattern: A pattern against which to match files on.
        excludes: A list of directory names to not look in.
        followlinks: Whether or not to follow symbolic links.

    Yields:
        File paths that match the filters provided.
    """
    if excludes is None:
        excludes = []

    for root, dirs, files in os.walk(directory, topdown=True, followlinks=followlinks):
        dirs[:] = [d for d in dirs if d not in excludes]
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = Path(root) / basename

                yield str(filename)


def render_template(template: str, **kwargs: str | dict[str, str]) -> str:
    """Render a jinaj2 template.

    Args:
        template: The jinja template to render.
        **kwargs: Values to render in the template.

    Returns:
        The rendered template.
    """
    t = jinja2.Environment(
        autoescape=jinja2.select_autoescape(
            default_for_string=False,
            default=False,
        ),
    )
    return t.from_string(template).render(kwargs)


def write_file(filename: str | Path, content: str, header: str | None = None) -> None:
    """Write a file with the given filename and content.

    Args:
        filename: The target file.
        content: A string containing the data to be written.
        header: A header, if None it will use default header.
    """
    if header is None:
        content = MOLECULE_HEADER + "\n\n" + content

    if isinstance(filename, str):
        filename = Path(filename)
    filename.parent.mkdir(exist_ok=True)
    filename.write_text(content)


def molecule_prepender(content: str) -> str:
    """Return molecule identification header.

    Args:
        content: Molecule content to prepend.

    Returns:
        Provided content prepended with MOLECULE_HEADER.
    """
    return MOLECULE_HEADER + "\n\n" + content


def file_prepender(filename: str | Path) -> None:
    """Prepend an informational header on files managed by Molecule and returns None.

    Args:
        filename: A string containing the target filename.
    """
    if isinstance(filename, str):
        filename = Path(filename)

    with filename.open("r+") as f:
        content = f.read()
        f.seek(0, 0)
        f.write(molecule_prepender(content))


def safe_dump(data: object, explicit_start: bool = True) -> str:  # noqa: FBT001, FBT002
    """Dump the provided data to a YAML document and returns a string.

    Args:
        data: The object to dump.
        explicit_start: An optional bool to toggle explicit document start.

    Returns:
        The YAML document in a string.
    """
    return yaml.dump(
        data,
        Dumper=SafeDumper,
        default_flow_style=False,
        explicit_start=explicit_start,
    )


def safe_load(string: str | TextIOWrapper):  # type: ignore[no-untyped-def]  # noqa: ANN201
    """Parse the provided string returns a dict.

    Args:
        string: A string to be parsed.

    Returns:
        A dict of the parsed string.


    Raises:
        MoleculeError: when YAML loading fails.
    """
    try:
        return yaml.safe_load(string) or {}
    except yaml.scanner.ScannerError as e:
        raise MoleculeError(str(e)) from None
    return {}


def safe_load_file(filename: str | Path):  # type: ignore[no-untyped-def]  # noqa: ANN201
    """Parse the provided YAML file and returns a dict.

    Args:
        filename: The file to parse.

    Returns:
        A dict of the parsed string.
    """
    if isinstance(filename, str):
        filename = Path(filename)

    with filename.open() as stream:
        return safe_load(stream)


def instance_with_scenario_name(instance_name: str, scenario_name: str) -> str:
    """Format instance name that includes scenario.

    Args:
        instance_name: Name of the instance.
        scenario_name: Name of the scenario.

    Returns:
        Combined instance and scenario names.
    """
    return f"{instance_name}-{scenario_name}"


def verbose_flag(options: Options) -> list[str]:
    """Return computed verbosity flag.

    Args:
        options: Full commandline options list.

    Returns:
        The appropriate verbosity option.
    """
    verbose = "v"
    flags = []
    for _i in range(3):
        if options.get(verbose):
            flags = [f"-{verbose}"]
            del options[verbose]
            if options.get("verbose"):
                del options["verbose"]
            break
        verbose = verbose + "v"

    return flags


def filter_verbose_permutation(options: Options) -> Options:
    """Clean verbose information.

    Args:
        options: Dictionary of commandline options.

    Returns:
        Dictionary of options without verbose options included.
    """
    return {k: options[k] for k in options if not re.match(r"^[v]+$", k)}


@overload
def abs_path(path: Path) -> Path: ...


@overload
def abs_path(path: str | None) -> str: ...


def abs_path(path: str | Path | None) -> str | Path:
    """Return absolute path.

    Args:
        path: File path to create absolute path from.

    Returns:
        Absolute path of path.
    """
    if not path:
        return ""

    output_type = type(path)
    if isinstance(path, Path):
        path = str(path)
    path = os.path.abspath(path)  # noqa: PTH100

    return output_type(path)


def merge_dicts(a: _T, b: _T) -> _T:
    """Merge the values of b into a and returns a new dict.

    This function uses the same algorithm as Ansible's `combine(recursive=True)` filter.

    Args:
        a: the target dictionary
        b: the dictionary to import

    Returns:
        A dictionary with b applied on to a
    """
    result = copy.deepcopy(a)

    for k, b_v in b.items():
        a_v = a.get(k)
        if a_v is not None and isinstance(a_v, dict) and isinstance(b_v, dict):
            result[k] = merge_dicts(a_v, b_v)
        else:
            result[k] = b_v

    return result


def validate_parallel_cmd_args(cmd_args: CommandArgs) -> None:
    """Prevents use of options incompatible with parallel mode.

    Args:
        cmd_args: Arguments to validate.

    Raises:
        MoleculeError: when incompatible options are present.
    """
    if cmd_args.get("parallel") and cmd_args.get("destroy") == "never":
        msg = 'Combining "--parallel" and "--destroy=never" is not supported'
        raise MoleculeError(msg)


def _parallelize_platforms(
    config: ConfigData,
    run_uuid: str,
) -> list[PlatformData]:
    def parallelize(platform: PlatformData) -> PlatformData:
        platform["name"] = f"{platform['name']}-{run_uuid}"
        return platform

    return [parallelize(platform) for platform in config["platforms"]]


def _filter_platforms(
    config: ConfigData,
    platform_name: str,
) -> list[PlatformData]:
    for platform in config["platforms"]:
        if platform["name"] == platform_name:
            return [platform]

    return []


@cache
def find_vcs_root(
    location: str | Path = "",
    dirs: tuple[str, ...] = (".git", ".hg", ".svn"),
    default: str = "",
) -> str:
    """Return current repository root directory.

    Args:
        location: Location to begin searching from.
        dirs: VCS config directory names to look for.
        default: Default value to return if no VCS directory found.

    Returns:
        The ancestor path containing VCS folders or default if none are found.
    """
    if not location:
        location = Path.cwd()
    if isinstance(location, str):
        location = Path(location)

    prev, location = None, location.absolute()
    while prev != location:
        if any((location / d).is_dir() for d in dirs):
            return str(location)
        prev, location = (location, location.parent)
    return default


def lookup_config_file(filename: str) -> str | None:
    """Return config file PATH.

    Args:
        filename: Config file name to find.and

    Returns:
        Path to config file or None if not found.
    """
    for path in [find_vcs_root(default="~"), "~"]:
        f = (Path(path) / filename).expanduser()
        if f.is_file():
            LOG.info("Found config file %s", f)
            return str(f)
    return None


def boolean(value: bool | AnyStr, *, strict: bool = True) -> bool:  # noqa: FBT001
    """Evaluate any object as boolean matching ansible behavior.

    Args:
        value: The value to evaluate as a boolean.
        strict: If True, invalid booleans will raises TypeError instead of returning False.

    Returns:
        The boolean value of value.

    Raises:
        TypeError: If value does not resolve to a valid boolean and strict is True.
    """
    # Based on https://github.com/ansible/ansible/blob/devel/lib/ansible/module_utils/parsing/convert_bool.py

    BOOLEANS_TRUE = frozenset(  # noqa: N806
        ("y", "yes", "on", "1", "true", "t", 1, 1.0, True),
    )
    BOOLEANS_FALSE = frozenset(  # noqa: N806
        ("n", "no", "off", "0", "false", "f", 0, 0.0, False, ""),
    )
    BOOLEANS = BOOLEANS_TRUE.union(BOOLEANS_FALSE)  # noqa: N806

    if isinstance(value, bool):
        return value

    normalized_value = str(value).lower().strip()

    if normalized_value in BOOLEANS_TRUE:
        return True
    if normalized_value in BOOLEANS_FALSE or not strict:
        return False

    raise TypeError(  # noqa: TRY003
        f"The value '{value!s}' is not a valid boolean.  Valid booleans include: {', '.join(repr(i) for i in BOOLEANS)!s}",  # noqa: EM102
    )


def dict2args(data: MutableMapping[str, str | bool]) -> list[str]:
    """Convert a dictionary of options to command like arguments.

    Args:
        data: Arguments dictionary to flatten.

    Returns:
        A list of command-like flags represented by the dictionary.
    """
    result = []
    # keep sorting in order to achieve a predictable behavior
    for k, v in sorted(data.items()):
        if v is not False:
            prefix = "-" if len(k) == 1 else "--"
            result.append(f"{prefix}{k}".replace("_", "-"))
            if v is not True:
                # { foo: True } should produce --foo without any values
                result.append(v)
    return result


def bool2args(data: bool | list[str]) -> list[str]:  # noqa: ARG001, FBT001
    """Convert a boolean value to command line argument (flag).

    Args:
        data: A boolean value.

    Returns:
        An empty list
    """
    return []


def print_as_yaml(data: object) -> None:
    """Render python object as yaml on console.

    Args:
        data: A YAML object.
    """
    # https://github.com/Textualize/rich/discussions/990#discussioncomment-342217
    result = Syntax(code=safe_dump(data), lexer="yaml", background_color="default")
    console.print(result)


def oxford_comma(listed: Iterable[bool | str | Path], condition: str = "and") -> str:
    """Format a list into a sentence.

    Args:
        listed: List of string entries to modify
        condition: String to splice into string, usually 'and'

    Returns:
        Modified string
    """
    match [f"'{entry!s}'" for entry in listed]:
        case [one]:
            return one
        case [one, two]:
            return f"{one} {condition} {two}"
        case [*front, back]:
            return f"{', '.join(s for s in front)}, {condition} {back}"
        case _:
            return ""



"""Base definition for Ansible Galaxy dependencies."""

from __future__ import annotations

from typing import TYPE_CHECKING

from molecule import util
from molecule.dependency.ansible_galaxy.collections import Collections
from molecule.dependency.ansible_galaxy.roles import Roles
from molecule.dependency.base import Base


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from molecule.config import Config


class AnsibleGalaxy(Base):
    """Galaxy is the default dependency manager.

    From v6.0.0, the dependencies are installed in the directory that defined
    in ansible configuration. The default installation directory is
    [DEFAULT_ROLES_PATH][]([ANSIBLE_HOME][]). If two versions of the same
    dependency is required, there is a conflict if the default installation
    directory is used because both are tried to be installed in one directory.

    Additional options can be passed to ``ansible-galaxy install`` through the
    options dict.  Any option set in this section will override the defaults.

    The `role-file` and `requirements-file` search path is `<role-name>`
    directory. The default value for `role-file` is `requirements.yml`, and the
    default value for `requirements-file` is `collections.yml`.

    1. If they are not defined in `options`, `molecule` will find them from the
    `<role-name>` directory, e.g. `<role-name>/requirements.yml` and
    `<role-name>/collections.yml`
    2. If they are defined in `options`, `molecule` will find them from
    `<role-name>/<the-value-of-role-file>` and
    `<role-name>/<the-value-of-requirements-file>`.

    !!! note

        Molecule will remove any options matching '^[v]+$', and pass ``-vvv``
        to the underlying ``ansible-galaxy`` command when executing
        `molecule --debug`.

    ``` yaml
        dependency:
          name: galaxy
          options:
            ignore-certs: True
            ignore-errors: True
            role-file: requirements.yml
            requirements-file: collections.yml
    ```

    Use "role-file" if you have roles only. Use the "requirements-file" if you
    need to install collections. Note that, with Ansible Galaxy's collections
    support, you can now combine the two lists into a single requirement if your
    file looks like this

    ``` yaml
        roles:
          - dep.role1
          - dep.role2
        collections:
          - ns.collection
          - ns2.collection2
    ```

    If you want to combine them, then just point your ``role-file`` and
    ``requirements-file`` to the same path. This is not done by default because
    older ``role-file`` only required a list of roles, while the collections
    must be under the ``collections:`` key within the file and pointing both to
    the same file by default could break existing code.

    The dependency manager can be disabled by setting ``enabled`` to False.

    ``` yaml
        dependency:
          name: galaxy
          enabled: False
    ```

    Environment variables can be passed to the dependency.

    ``` yaml
        dependency:
          name: galaxy
          env:
            FOO: bar
    ```

    [DEFAULT_ROLES_PATH]: https://docs.ansible.com/ansible/latest/cli/ansible-galaxy.html#cmdoption-ansible-galaxy-role-remove-p
    [ANSIBLE_HOME]: https://docs.ansible.com/ansible/latest/reference_appendices/config.html#ansible-home
    """

    def __init__(self, config: Config) -> None:
        """Construct AnsibleGalaxy.

        Args:
            config: Molecule Config instance.
        """
        super().__init__(config)
        self.invocations = [Roles(config), Collections(config)]

    def execute(self, action_args: list[str] | None = None) -> None:  # noqa: ARG002
        """Execute all Ansible Galaxy dependencies.

        Args:
            action_args: Arguments for invokers. Unused.
        """
        for invoker in self.invocations:
            invoker.execute()

    def _has_requirements_file(self) -> bool:
        has_file = False
        for invoker in self.invocations:
            has_file = has_file or invoker._has_requirements_file()  # noqa: SLF001
        return has_file

    @property
    def default_env(self) -> dict[str, str]:
        """Default environment variables across all invokers.

        Returns:
            Merged dictionary of default env vars for all invokers.
        """
        env: dict[str, str] = {}
        for invoker in self.invocations:
            env = util.merge_dicts(env, invoker.default_env)
        return env

    @property
    def default_options(self) -> MutableMapping[str, str | bool]:
        """Default options across all invokers.

        Returns:
            Merged dictionary of default options for all invokers.
        """
        opts: MutableMapping[str, str | bool] = {}
        for invoker in self.invocations:
            opts = util.merge_dicts(opts, invoker.default_options)
        return opts



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Molecule Shell Module."""

from __future__ import annotations

import atexit
import os
import sys

from pathlib import Path

import click
import packaging

import molecule

from molecule import command, logger
from molecule.api import drivers
from molecule.app import get_app
from molecule.click_cfg import click_group_ex
from molecule.config import MOLECULE_DEBUG, MOLECULE_VERBOSITY
from molecule.console import console
from molecule.util import do_report, lookup_config_file


# Setup logging. This location of initialization is not ideal, but the code
# structure does not give us much choice because config file lookup down below
# uses logging facilities. Do note that verbosity level set by the
# command-line flags does not affect the things that execute before the main
# method in this file that parses the command line arguments.
#
# It would be ideal if we could get a chance at parsing CLI args before any of
# the real code executes, but this is really hard to do while using click for
# constructing out CLI API.
logger.configure()

LOCAL_CONFIG_SEARCH = ".config/molecule/config.yml"
LOCAL_CONFIG = lookup_config_file(LOCAL_CONFIG_SEARCH)

ENV_FILE = ".env.yml"


def print_version(
    ctx: click.Context,
    param: click.Option,  # noqa: ARG001
    value: bool,  # noqa: FBT001
) -> None:
    """Print version information.

    Args:
        ctx: Click context object.
        param: Reference to the --version option.
        value: Value of --version.
    """
    if not value or ctx.resilient_parsing:
        return

    v = packaging.version.Version(molecule.__version__)
    color = "bright_yellow" if v.is_prerelease else "green"
    msg = (
        f"molecule [{color}]{v}[/] "
        f"using python [repr.number]{sys.version_info[0]}.{sys.version_info[1]}[/] \n"
    )

    msg += f"    [repr.attrib_name]ansible[/][dim]:[/][repr.number]{get_app(Path()).runtime.version}[/]"
    for driver in drivers().values():
        msg += (
            f"\n    [repr.attrib_name]{driver!s}[/][dim]:[/][repr.number]{driver.version}[/][dim] "
            f"from {driver.module}"
        )
        if driver.required_collections:
            msg += " requiring collections:"
            for name, version in driver.required_collections.items():
                msg += f" [repr.attrib_name]{name}[/]>=[repr.number]{version}[/]"
        msg += "[/]"
    console.print(msg, highlight=False)

    ctx.exit()


@click_group_ex()
@click.option(
    "--debug/--no-debug",
    default=MOLECULE_DEBUG,
    help="Enable or disable debug mode. Default is disabled.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    default=MOLECULE_VERBOSITY,
    help="Increase Ansible verbosity level. Default is 0.",
)
@click.option(
    "--base-config",
    "-c",
    multiple=True,
    default=[LOCAL_CONFIG] if LOCAL_CONFIG else [],
    help=(
        "Path to a base config (can be specified multiple times)."
        " If provided, Molecule will first load and deep merge the"
        " configurations in the specified order,"
        " and deep merge each scenario's "
        "molecule.yml on top. By default Molecule is looking for "
        f"'{LOCAL_CONFIG_SEARCH}' "
        "in current VCS repository and if not found it will look "
        f"in user home. ({LOCAL_CONFIG})."
    ),
)
@click.option(
    "--env-file",
    "-e",
    default=ENV_FILE,
    help=("The file to read variables from when rendering molecule.yml. (.env.yml)"),
)
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Show version information and exit.",
)
@click.pass_context
def main(
    ctx: click.Context,
    debug: bool,  # noqa: FBT001
    verbose: int,
    base_config: list[str],
    env_file: str,
) -> None:  # pragma: no cover
    """Molecule aids in the development and testing of Ansible roles.

    To enable autocomplete for a supported shell execute command below after
    replacing SHELL with either bash, zsh, or fish:

        eval "$(_MOLECULE_COMPLETE=SHELL_source molecule)"

    Args:
        ctx: Click context object.
        debug: Debug option value.
        verbose: Verbose option value.
        base_config: Base config option value.
        env_file: Environment variable file option value.
    """
    ctx.obj = {}
    ctx.obj["args"] = {}
    ctx.obj["args"]["debug"] = debug
    ctx.obj["args"]["verbose"] = verbose
    ctx.obj["args"]["base_config"] = base_config
    ctx.obj["args"]["env_file"] = env_file

    logger.set_log_level(verbose, debug)
    if verbose:
        os.environ["ANSIBLE_VERBOSITY"] = str(verbose)

    if "MOLECULE_REPORT" in os.environ:
        atexit.register(do_report)


main.add_command(command.cleanup.cleanup)
main.add_command(command.check.check)
main.add_command(command.converge.converge)
main.add_command(command.create.create)
main.add_command(command.dependency.dependency)
main.add_command(command.destroy.destroy)
main.add_command(command.drivers.drivers)
main.add_command(command.idempotence.idempotence)
main.add_command(command.init.init)
main.add_command(command.list.list_)
main.add_command(command.login.login)
main.add_command(command.matrix.matrix)
main.add_command(command.prepare.prepare)
main.add_command(command.reset.reset)
main.add_command(command.side_effect.side_effect)
main.add_command(command.syntax.syntax)
main.add_command(command.test.test)
main.add_command(command.verify.verify)



# Copyright 2015 Docker, Inc.  # noqa: D100
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import pytest

from molecule import interpolation


@pytest.fixture
def _mock_env():  # type: ignore[no-untyped-def]  # noqa: ANN202
    return {
        "FOO": "foo",
        "BAR": "",
        "DEPENDENCY_NAME": "galaxy",
        "VERIFIER_NAME": "ansible",
        "MOLECULE_SCENARIO_NAME": "default",
    }


@pytest.fixture
def _instance(_mock_env):  # type: ignore[no-untyped-def]  # noqa: ANN202
    return interpolation.Interpolator(interpolation.TemplateWithDefaults, _mock_env)


def test_escaped_interpolation(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.interpolate("$${foo}") == "${foo}"


def test_invalid_interpolation(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    with pytest.raises(interpolation.InvalidInterpolation):
        _instance.interpolate("${")
    with pytest.raises(interpolation.InvalidInterpolation):
        _instance.interpolate("$}")
    with pytest.raises(interpolation.InvalidInterpolation):
        _instance.interpolate("${}")
    with pytest.raises(interpolation.InvalidInterpolation):
        _instance.interpolate("${ }")
    with pytest.raises(interpolation.InvalidInterpolation):
        _instance.interpolate("${ foo}")
    with pytest.raises(interpolation.InvalidInterpolation):
        _instance.interpolate("${foo }")
    with pytest.raises(interpolation.InvalidInterpolation):
        _instance.interpolate("${foo!}")


def test_interpolate_missing_no_default(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.interpolate("This ${missing} var") == "This  var"
    assert _instance.interpolate("This ${BAR} var") == "This  var"


def test_interpolate_with_value(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.interpolate("This $FOO var") == "This foo var"
    assert _instance.interpolate("This ${FOO} var") == "This foo var"


def test_interpolate_missing_with_default(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.interpolate("ok ${missing:-def}") == "ok def"
    assert _instance.interpolate("ok ${missing-def}") == "ok def"
    assert (
        _instance.interpolate(
            "ok ${BAR:-/non:-alphanumeric}",
        )
        == "ok /non:-alphanumeric"
    )


def test_interpolate_with_empty_and_default_value(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    assert _instance.interpolate("ok ${BAR:-def}") == "ok def"
    assert _instance.interpolate("ok ${BAR-def}") == "ok "


def test_interpolate_interpolates_MOLECULE_strings(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, N802, D103
    assert _instance.interpolate("$MOLECULE_SCENARIO_NAME") == "default"


def test_interpolate_does_not_interpolate_MOLECULE_strings(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, N802, D103
    assert (
        _instance.interpolate(
            "foo $MOLECULE_SCENARIO_NAME",
            keep_string="MOLECULE_",
        )
        == "foo $MOLECULE_SCENARIO_NAME"
    )


def test_interpolate_with_molecule_yaml(_instance):  # type: ignore[no-untyped-def]  # noqa: ANN201, PT019, D103
    data = """
---
dependency:
    name: $DEPENDENCY_NAME
driver:
    name: default
platforms:
  - name: instance-1
provisioner:
    name: ansible
verifier:
    name: ${VERIFIER_NAME}
    options:
      $FOO: bar
""".strip()

    x = """
---
dependency:
    name: galaxy
driver:
    name: default
platforms:
  - name: instance-1
provisioner:
    name: ansible
verifier:
    name: ansible
    options:
      foo: bar
""".strip()

    assert x == _instance.interpolate(data)



"""New Click Configuration System with CliOption Architecture.

This module provides a framework-agnostic approach to CLI option management
that supports both Click and future migration to argparse or other frameworks.
It also contains Click-specific decorators and utilities.
"""

from __future__ import annotations

import functools

from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, Any, NoReturn

import click

from click_help_colors import HelpColorsCommand, HelpColorsGroup

from molecule import util
from molecule.ansi_output import should_do_markup
from molecule.api import drivers
from molecule.config import MOLECULE_PARALLEL
from molecule.constants import MOLECULE_DEFAULT_SCENARIO_NAME, MOLECULE_PLATFORM_NAME


if TYPE_CHECKING:
    from collections.abc import Callable

    ClickCommand = Callable[[Callable[..., None]], click.Command]
    ClickGroup = Callable[[Callable[..., None]], click.Group]


# Common option combinations for different command types
# Options that most workflow commands share
COMMON_OPTIONS = [
    "all_scenarios",
    "exclude",
    "report",
    "scenario_name_with_default",
    "shared_inventory",
    "shared_state",
]


@dataclass
class CliOption:
    """Framework-agnostic CLI option definition.

    Attributes:
        name: The name of the CLI option.
        help: Help text for the option.
        short: Short form flag (e.g., "-s") or None.
        multiple: Whether the option can be specified multiple times.
        default: Default value for the option.
        type: Python type for the option value.
        choices: List of valid choices for the option.
        is_flag: Whether this is a boolean flag option.
        required: Whether the option is required.
        is_argument: Whether this is a positional argument rather than an option.
        nargs: Number of arguments for positional arguments.
        experimental: Whether this is an experimental feature.
        help_default: Custom default value to display in help text.
    """

    name: str
    help: str
    short: str | None = None
    multiple: bool = False
    default: Any = None
    type: type | None = None
    choices: list[str] | None = None
    is_flag: bool = False
    required: bool = False
    is_argument: bool = False
    nargs: int = 1
    experimental: bool = False
    help_default: str | None = None

    def _generate_help_text(self) -> str:
        """Generate help text with automatic default value inclusion.

        Returns:
            Formatted help text with experimental prefix and default info.
        """
        help_text = self.help

        # Add experimental prefix
        if self.experimental:
            help_text = f"EXPERIMENTAL: {help_text}"

        # Early exit for arguments - they don't show default info
        if self.is_argument:
            return help_text

        default_value = None

        # Use custom help_default if provided
        if self.help_default is not None:
            default_value = self.help_default
        # For flags, show disabled/enabled
        elif self.is_flag and self.default is not None:
            default_value = "enabled" if self.default else "disabled"
        # For options with actual defaults
        elif self.default is not None:
            if self.multiple and isinstance(self.default, list) and self.default:
                if len(self.default) == 1:
                    default_value = str(self.default[0])
                else:
                    default_value = ", ".join(map(str, self.default))
            else:
                default_value = str(self.default)

        # Append default info if we have a value
        if default_value is not None:
            help_text += f" (default: {default_value})"

        return help_text

    def as_click_option(self) -> Callable[..., Any]:
        """Convert to Click option decorator.

        Returns:
            A Click option or argument decorator function.
        """
        if self.is_argument:
            return click.argument(
                self.name.replace("-", "_"),
                nargs=self.nargs,
                type=click.UNPROCESSED if self.type is None else self.type,
            )

        params = [f"--{self.name}"]
        if self.short:
            params.append(self.short)

        kwargs = {
            "help": self._generate_help_text(),
            "default": self.default,
            "multiple": self.multiple,
            "required": self.required,
        }

        # Type and choices
        if self.choices:
            kwargs["type"] = click.Choice(self.choices)
        elif self.type:
            kwargs["type"] = self.type

        # Handle flag options
        if self.is_flag:
            return click.option(f"--{self.name}/--no-{self.name}", **kwargs)

        return click.option(*params, **kwargs)


class CliOptions:
    """Central registry of all CLI options with support for variants."""

    # Alphabetically ordered properties

    @property
    def all_scenarios(self) -> CliOption:
        """Target all scenarios option."""
        return CliOption(
            name="all",
            help="Target all scenarios. Overrides scenario-name.",
            is_flag=True,
            default=False,
        )

    @property
    def ansible_args(self) -> CliOption:
        """Ansible arguments."""
        return CliOption(
            name="ansible_args",
            help="Arguments to forward to Ansible",
            is_argument=True,
            nargs=-1,
        )

    @property
    def dependency_name(self) -> CliOption:
        """Dependency name option."""
        return CliOption(
            name="dependency-name",
            help="Name of dependency to initialize.",
            choices=["galaxy"],
            default="galaxy",
        )

    @property
    def destroy(self) -> CliOption:
        """Destroy strategy option."""
        return CliOption(
            name="destroy",
            help="The destroy strategy used at the conclusion of a Molecule run.",
            choices=["always", "never"],
            default="always",
        )

    @property
    def driver_name(self) -> CliOption:
        """Driver name option without choices."""
        return CliOption(
            name="driver-name",
            help="Name of the driver to use.",
            short="-d",
        )

    @property
    def driver_name_with_choices(self) -> CliOption:
        """Driver name option with available choices."""
        return replace(
            self.driver_name,
            choices=[str(s) for s in drivers()],
            help="Name of driver to use.",
        )

    @property
    def exclude(self) -> CliOption:
        """Exclude scenarios option."""
        return CliOption(
            name="exclude",
            help="Name of the scenario to exclude from targeting. May be specified multiple times.",
            short="-e",
            multiple=True,
        )

    @property
    def force(self) -> CliOption:
        """Force execution option."""
        return CliOption(
            name="force",
            help="Enable or disable force mode.",
            short="-f",
            is_flag=True,
            default=False,
        )

    @property
    def format_full(self) -> CliOption:
        """Format option with full choices including yaml."""
        return CliOption(
            name="format",
            help="Change output format.",
            short="-f",
            choices=["simple", "plain", "yaml"],
            default="simple",
        )

    @property
    def format_simple(self) -> CliOption:
        """Format option with simple choices."""
        return CliOption(
            name="format",
            help="Change output format.",
            short="-f",
            choices=["simple", "plain"],
            default="simple",
        )

    @property
    def host(self) -> CliOption:
        """Host connection option."""
        return CliOption(
            name="host",
            help="Host to access.",
            short="-h",
        )

    @property
    def parallel(self) -> CliOption:
        """Parallel execution option."""
        return CliOption(
            name="parallel",
            help="Enable or disable parallel mode.",
            is_flag=True,
            default=MOLECULE_PARALLEL,
        )

    @property
    def platform_name(self) -> CliOption:
        """Platform name option."""
        return CliOption(
            name="platform-name",
            help="Name of the platform to target.",
            short="-p",
        )

    @property
    def platform_name_with_default(self) -> CliOption:
        """Platform name option with default."""
        return replace(
            self.platform_name,
            default=MOLECULE_PLATFORM_NAME,
            help="Name of the platform to target only.",
            help_default="None",
        )

    @property
    def provisioner_name(self) -> CliOption:
        """Provisioner name option."""
        return CliOption(
            name="provisioner-name",
            help="Name of provisioner to initialize.",
            choices=["ansible"],
            default="ansible",
        )

    @property
    def report(self) -> CliOption:
        """Reporting option."""
        return CliOption(
            name="report",
            help="Enable or disable end-of-run summary report.",
            is_flag=True,
            default=False,
            experimental=True,
        )

    @property
    def scenario_name(self) -> CliOption:
        """Base scenario name option without default."""
        return CliOption(
            name="scenario-name",
            help="Name of the scenario to target.  May be specified multiple times.",
            short="-s",
            multiple=True,
        )

    @property
    def scenario_name_with_default(self) -> CliOption:
        """Scenario name option with default value (multiple)."""
        return replace(
            self.scenario_name,
            default=[MOLECULE_DEFAULT_SCENARIO_NAME],
        )

    @property
    def scenario_name_single(self) -> CliOption:
        """Single scenario name option without default."""
        return replace(
            self.scenario_name,
            multiple=False,
            help="Name of the scenario to target.",
        )

    @property
    def scenario_name_single_with_default(self) -> CliOption:
        """Single scenario name option with default."""
        return replace(
            self.scenario_name_single,
            default=MOLECULE_DEFAULT_SCENARIO_NAME,
            help="Name of the scenario to target.",
        )

    @property
    def shared_inventory(self) -> CliOption:
        """Shared inventory option."""
        return CliOption(
            name="shared-inventory",
            help="Enable or disable sharing inventory between scenarios.",
            is_flag=True,
            default=False,
            experimental=True,
        )

    @property
    def shared_state(self) -> CliOption:
        """Shared state option."""
        return CliOption(
            name="shared-state",
            help="Enable or disable sharing (some) state between scenarios.",
            is_flag=True,
            default=False,
            experimental=True,
        )

    @property
    def subcommand(self) -> CliOption:
        """Subcommand argument for matrix command."""
        return CliOption(
            name="subcommand",
            help="Subcommand to analyze",
            is_argument=True,
            nargs=1,
        )


def options(option_names: list[str]) -> Callable[..., Any]:
    """Decorator that adds CLI options to a command function.

    Args:
        option_names: List of option names to apply from CliOptions class.

    Example:
        @options(["scenario_name_with_default", "parallel", "force"])
        def my_command(ctx: click.Context) -> None:
            scenario = ctx.params["scenario_name"]
            parallel = ctx.params["parallel"]
            force = ctx.params["force"]

    Returns:
        The decorated function with the specified Click options applied.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        cli_options = CliOptions()

        # Create the wrapper function with click.pass_context applied to IT
        @click.pass_context  # type: ignore[arg-type]
        @functools.wraps(func)
        def wrapper(ctx: click.Context, *_args: object, **_kwargs: object) -> object:
            # All options and arguments are available via ctx.params
            # Functions only need ctx, so we ignore any additional args/kwargs Click might pass
            return func(ctx)

        # Get all the CliOption instances
        options_list = [getattr(cli_options, option_name) for option_name in option_names]

        # Sort options and apply them to the wrapper (reverse order for correct display)
        sorted_options = _sort_options(options_list)
        for option in reversed(sorted_options):
            click_option = option.as_click_option()
            wrapper = click_option(wrapper)

        return wrapper

    return decorator


def common_options(*additional_options: str) -> Callable[..., Any]:
    """Decorator that adds common CLI options plus any additional options to a command function.

    This decorator includes COMMON_OPTIONS by default and adds any additional
    options provided as arguments.

    Args:
        *additional_options: Additional option names to include beyond COMMON_OPTIONS.

    Example:
        @common_options("parallel", "force")
        def my_command(ctx: click.Context) -> None:
            # All COMMON_OPTIONS plus parallel and force are available
            scenario = ctx.params["scenario_name"]  # from COMMON_OPTIONS
            parallel = ctx.params["parallel"]       # additional
            force = ctx.params["force"]             # additional

        @common_options()
        def my_other_command(ctx: click.Context) -> None:
            # Just COMMON_OPTIONS are available
            scenario = ctx.params["scenario_name"]

    Returns:
        The decorated function with common options and additional options applied.
    """
    option_names = list(set(COMMON_OPTIONS + list(additional_options)))
    return options(option_names)


def _sort_options(option_list: list[CliOption]) -> list[CliOption]:
    """Sort options in user-friendly order with clear sections.

    Order:
    1. Core workflow options (scenario-name, exclude, all)
    2. Options with short forms (alphabetical)
    3. Long options without short forms (alphabetical)
    4. Experimental options (alphabetical)

    Args:
        option_list: List of CliOption instances.

    Returns:
        List of sorted CliOption instances.
    """
    # Convert to dict for easy lookup and removal
    options_by_name = {opt.name: opt for opt in option_list}

    # Core workflow options (scenario-name, exclude, all)
    core_workflow_names = ["scenario-name", "exclude", "all"]
    core_workflow_options = [
        options_by_name.pop(name) for name in core_workflow_names if name in options_by_name
    ]

    # Extract and sort each category
    experimental_options = sorted(
        [opt for opt in options_by_name.values() if opt.experimental],
        key=lambda opt: opt.name,
    )
    short_form_options = sorted(
        [opt for opt in options_by_name.values() if not opt.experimental and opt.short is not None],
        key=lambda opt: opt.name,
    )
    long_form_options = sorted(
        [opt for opt in options_by_name.values() if not opt.experimental and opt.short is None],
        key=lambda opt: opt.name,
    )

    # Build final list by combining all sections
    return core_workflow_options + short_form_options + long_form_options + experimental_options


class FirstLineHelpMixin:
    """Mixin to modify help text to only show the first line.

    Attributes:
        help: The help text that will be modified to show only first line.
    """

    help: str | None

    def __init__(self, *args: object, **kwargs: object) -> None:
        """Initialize the command.

        Args:
            *args: Positional arguments to pass to parent.
            **kwargs: Keyword arguments to pass to parent.
        """
        super().__init__(*args, **kwargs)
        if self.help:
            first_line = self.help.strip().split("\n")[0].strip()
            self.help = first_line


class FirstLineHelpCommand(FirstLineHelpMixin, HelpColorsCommand):
    """Override the default help generation to remove the args section."""


class FirstLineHelpGroup(FirstLineHelpMixin, HelpColorsGroup):
    """Override the default help generation to remove the args section."""


def click_group_ex() -> ClickGroup:
    """Return extended version of click.group().

    Returns:
        Click command group.
    """
    # Color coding used to group command types, documented only here as we may
    # decide to change them later.
    # green : (default) as sequence step
    # blue : molecule own command, not dependent on scenario
    # yellow : special commands, like full test sequence, or login
    return click.group(
        cls=FirstLineHelpGroup,
        # Workaround to disable click help line truncation to ~80 chars
        # https://github.com/pallets/click/issues/486
        context_settings={
            "max_content_width": 9999,
            "color": should_do_markup(),
            "help_option_names": ["-h", "--help"],
        },
        help_headers_color="yellow",
        help_options_color="green",
        help_options_custom_colors={
            "drivers": "blue",
            "init": "blue",
            "list": "blue",
            "matrix": "blue",
            "login": "bright_yellow",
            "reset": "blue",
            "test": "bright_yellow",
        },
        result_callback=result_callback,
    )


def click_command_ex(name: str | None = None) -> ClickCommand:
    """Return extended version of click.command().

    Args:
        name: A replacement name in the case the automatic one is insufficient.

    Returns:
        Click command group.
    """
    return click.command(
        cls=FirstLineHelpCommand,
        name=name,
        help_headers_color="yellow",
        help_options_color="green",
    )


def result_callback(
    *args: object,  # noqa: ARG001
    **kwargs: object,  # noqa: ARG001
) -> NoReturn:
    """Click natural exit callback.

    Args:
        *args: Unused.
        **kwargs: Unused.
    """
    # We want to be used we run out custom exit code, regardless if run was
    # a success or failure.
    util.sysexit(0)



#  Copyright (c) 2015-2018 Cisco Systems, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
"""Molecule Scenario Module."""

from __future__ import annotations

import fcntl
import fnmatch
import os
import shutil

from functools import cached_property
from pathlib import Path
from time import sleep
from typing import TYPE_CHECKING

from molecule import logger, scenarios, util
from molecule.constants import RC_TIMEOUT
from molecule.exceptions import MoleculeError
from molecule.text import checksum


if TYPE_CHECKING:
    from molecule.config import Config
    from molecule.types import ScenarioResult


class Scenario:
    """A Molecule scenario."""

    def __init__(self, config: Config) -> None:
        """Initialize a new scenario class and returns None.

        Args:
            config: An instance of a Molecule config.
        """
        self._lock = None
        self.config = config
        self.results: list[ScenarioResult] = []
        self._setup()

    def __repr__(self) -> str:
        """Return a representation of the instance.

        Returns:
            A string.
        """
        return f"<Scenario {self.name} from {self.config.project_directory}>"

    def _remove_scenario_state_directory(self) -> None:
        """Remove scenario cached disk stored state."""
        directory = str(Path(self.ephemeral_directory).parent)
        scenario_log = logger.get_scenario_logger(__name__, self.name, "scenario")
        scenario_log.info("Removing %s", directory)
        shutil.rmtree(directory)

    def prune(self) -> None:
        """Prune the scenario ephemeral directory files.

        "safe files" will not be pruned, including the ansible configuration
        and inventory used by this scenario, the scenario state file, and
        files declared as "safe_files" in the ``driver`` configuration
        declared in ``molecule.yml``.
        """
        scenario_log = logger.get_scenario_logger(__name__, self.name, "scenario")
        scenario_log.info("Pruning extra files from scenario ephemeral directory")

        safe_files = [
            self.config.state.state_file,
            *self.config.driver.safe_files,
        ]
        if self.config.provisioner is not None:
            safe_files += [
                self.config.provisioner.config_file,
                self.config.provisioner.inventory_file,
            ]

        existing_files = util.os_walk(self.ephemeral_directory, "*")
        for f in existing_files:
            if not any(sf for sf in safe_files if fnmatch.fnmatch(f, sf)):
                Path(f).unlink(missing_ok=True)

        # Remove empty directories.
        for dirpath, dirs, files in os.walk(self.ephemeral_directory, topdown=False):
            if not dirs and not files:
                os.removedirs(dirpath)

    @property
    def name(self) -> str:
        """Name of the scenario.

        Returns:
            The scenario's name.
        """
        return self.config.config["scenario"]["name"]

    @property
    def directory(self) -> str:
        """Scenario directory.

        Returns:
            The directory containing the scenario.
        """
        path = Path.cwd()
        if self.config.molecule_file:
            path = Path(self.config.molecule_file).parent
        return str(path)

    @cached_property
    def ephemeral_directory(self) -> str:
        """Acquire the ephemeral directory.

        Returns:
            The ephemeral directory for this scenario.

        Raises:
            MoleculeError: If lock cannot be acquired before timeout.
        """
        path: Path
        if "MOLECULE_EPHEMERAL_DIRECTORY" not in os.environ:
            project_directory = Path(self.config.project_directory).name

            if self.config.is_parallel:
                project_directory = f"{project_directory}-{self.config._run_uuid}"  # noqa: SLF001

            project_scenario_directory = f"molecule.{checksum(project_directory, 4)}.{self.name}"
            path = self.config.runtime.cache_dir / "tmp" / project_scenario_directory
        else:
            path = Path(os.getenv("MOLECULE_EPHEMERAL_DIRECTORY", ""))
        path.mkdir(parents=True, exist_ok=True)

        if self.config.is_parallel and not self._lock:
            lock_file = path / ".lock"
            with lock_file.open("w") as self._lock:  # type: ignore[assignment]
                for i in range(1, 5):
                    try:
                        fcntl.lockf(self._lock, fcntl.LOCK_EX | fcntl.LOCK_NB)  # type: ignore[arg-type]
                        break
                    except OSError:
                        delay = 30 * i
                        scenario_log = logger.get_scenario_logger(__name__, self.name, "scenario")
                        scenario_log.warning(
                            "Retrying to acquire lock on %s, waiting for %s seconds",
                            path,
                            delay,
                        )
                        sleep(delay)
                else:
                    scenario_log = logger.get_scenario_logger(__name__, self.name, "scenario")
                    scenario_log.warning("Timedout trying to acquire lock on %s", path)
                    raise MoleculeError(code=RC_TIMEOUT)

        return path.absolute().as_posix()

    @cached_property
    def shared_ephemeral_directory(self) -> str:
        """Acquire the shared ephemeral directory.

        Returns:
            The common ephemeral directory for all scenarios.
        """
        path: Path
        if "MOLECULE_EPHEMERAL_DIRECTORY" not in os.environ:
            project_directory = Path(self.config.project_directory).name

            project_scenario_directory = f"molecule.{checksum(project_directory, 4)}"
            path = self.config.runtime.cache_dir / "tmp" / project_scenario_directory
        else:
            path = Path(os.getenv("MOLECULE_EPHEMERAL_DIRECTORY", ""))

        return path.absolute().as_posix()

    @property
    def inventory_directory(self) -> str:
        """Inventory directory.

        Returns:
            The directory containing the scenario's inventory.
        """
        if self.config.shared_inventory and not self.config.is_parallel:
            ephemeral = Path(self.shared_ephemeral_directory)
        else:
            ephemeral = Path(self.ephemeral_directory)

        return str(ephemeral / "inventory")

    @property
    def check_sequence(self) -> list[str]:
        """Check playbook sequence.

        Returns:
            A list of playbooks to run for 'check'.
        """
        return self.config.config["scenario"]["check_sequence"]

    @property
    def cleanup_sequence(self) -> list[str]:
        """Cleanup playbook sequence.

        Returns:
            A list of playbooks to run for 'cleanup'.
        """
        return self.config.config["scenario"]["cleanup_sequence"]

    @property
    def converge_sequence(self) -> list[str]:
        """Converge playbook sequence.

        Returns:
            A list of playbooks to run for 'converge'.
        """
        return self.config.config["scenario"]["converge_sequence"]

    @property
    def create_sequence(self) -> list[str]:
        """Create playbook sequence.

        Returns:
            A list of playbooks to run for 'create'.
        """
        return self.config.config["scenario"]["create_sequence"]

    @property
    def dependency_sequence(self) -> list[str]:
        """Dependency playbook sequence.

        Returns:
            A list of playbooks to run for 'dependency'.
        """
        return ["dependency"]

    @property
    def destroy_sequence(self) -> list[str]:
        """Destroy playbook sequence.

        Returns:
            A list of playbooks to run for 'destroy'.
        """
        return self.config.config["scenario"]["destroy_sequence"]

    @property
    def idempotence_sequence(self) -> list[str]:
        """Idempotence playbook sequence.

        Returns:
            A list of playbooks to run for 'idempotence'.
        """
        return ["idempotence"]

    @property
    def prepare_sequence(self) -> list[str]:
        """Prepare playbook sequence.

        Returns:
            A list of playbooks to run for 'prepare'.
        """
        return ["prepare"]

    @property
    def side_effect_sequence(self) -> list[str]:
        """Side effect playbook sequence.

        Returns:
            A list of playbooks to run for 'side_effect'.
        """
        return ["side_effect"]

    @property
    def syntax_sequence(self) -> list[str]:
        """Syntax playbook sequence.

        Returns:
            A list of playbooks to run for 'syntax'.
        """
        return ["syntax"]

    @property
    def test_sequence(self) -> list[str]:
        """Test playbook sequence.

        Returns:
            A list of playbooks to run for 'test'.
        """
        return self.config.config["scenario"]["test_sequence"]

    @property
    def verify_sequence(self) -> list[str]:
        """Verify playbook sequence.

        Returns:
            A list of playbooks to run for 'verify'.
        """
        return ["verify"]

    @property
    def sequence(self) -> list[str]:
        """Select the sequence based on scenario and subcommand of the provided scenario.

        Returns:
            A list of playbooks to run.

        Raises:
            RuntimeError: if an unexpected sequence type is requested.
        """
        result = []
        our_scenarios = scenarios.Scenarios([self.config])
        matrix = our_scenarios._get_matrix()  # noqa: SLF001

        try:
            result = matrix[self.name][self.config.subcommand]
            if not isinstance(result, list):
                raise RuntimeError(  # noqa: TRY003, TRY004
                    "Unexpected sequence type {result}.",  # noqa: EM101
                )
        except KeyError:
            pass
        return result

    def _setup(self) -> None:
        """Prepare the scenario for Molecule."""
        inventory = Path(self.inventory_directory)
        if not inventory.is_dir():
            inventory.mkdir(exist_ok=True, parents=True)
